import pandas as pd
from imblearn.under_sampling import RandomUnderSampler
from sklearn import dummy
from sklearn.feature_selection import RFECV, mutual_info_classif
from sklearn.impute import SimpleImputer
from sklearn.metrics import confusion_matrix, accuracy_score, f1_score, matthews_corrcoef
from sklearn.model_selection import StratifiedKFold, StratifiedShuffleSplit
from sklearn.preprocessing import LabelBinarizer, OneHotEncoder
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn_pandas import DataFrameMapper, gen_features

from plots.plots_benchmark_mode import plot_metrics
from plots.plots_machine_learning import *
from preprocessing import utils_tools


class Metapredictions(object):
    """
    Class that uses machine learning techniques to evaluate prediction tools.
    """

    def __init__(self, df: pd.DataFrame,
                 location: str,
                 thresholds: list,
                 out_dir: str,
                 top_tools: dict,
                 missing_rate_allowed: float = 0.3,
                 rank_metric: str = "weighted_accuracy",
                 undersample: bool = False,
                 only_numeric: bool = True):
        """
        :param pd.DataFrame df: Input data
        :param str location: Location of the variants associated
            with the input dataframe
        :param list thresholds: Reference thresholds of the tools
        :param str out_dir: Ouptut dir
        :param dict top_tools: Top tools generated by the performance
            comparison step (per each location), ranked by the
            metric provided by the user.
        :param float missing_rate_allowed: Fraction of missing values
            allowed per class in a tool so that it is not discarded from the
            analysis. Default: `0.3`. Tools with bigger rate will
            not be used. If missing rate is lower than this value,
            tool will be included and missing values will be imputed
            with median.
        :param str rank_metric: Metric to rank classifier performance
            to compare with individual scores provided in `top_tools`
            arg. Default: "weighted_accuracy".
        :param bool undersample: Undersample data based on majority class
        (useful for unbalanced data)
        :param bool only_numeric: Include only numerical features (tool scores).
            Default: `False`. If `True`, categorical and boolean features
            will not be included.
        """
        # Remove useless columns
        useless_cols = [c for c in ['index', 'chr', 'pos', 'ref', 'alt', 'id', 'HGVSc'] if c in df.columns]
        
        df = df.drop(useless_cols, axis=1)

        self.label_fields = ['label']  # 'outcome' would also work
        self.categorical_fields = ['type', 'subtype', 'Gene', 'Consequence']

        self.out_dir = out_dir
        self.location = location
        self.top_tools = top_tools
        self.rank_metric = rank_metric
        self.only_numeric = only_numeric

        # Remove tools with too many missing data in any of the classes
        _patho = df[df.label].copy()
        _benign = df[~df.label].copy()
        
        na_patho = set(list(df)).difference(set(list(_patho.dropna(thresh=_patho.shape[0] * missing_rate_allowed, axis=1))))
        na_benign = set(list(df)).difference(set(list(_benign.dropna(thresh=_benign.shape[0] * missing_rate_allowed, axis=1))))
        to_remove = list(na_patho.union(na_benign))
        logging.info('Features (tool scores) that will not be used because they display high rate of missing predictions (> {}): {}'.format(missing_rate_allowed,
                                                                                                                                           to_remove))
        
        self.df = df.drop(to_remove, axis=1)
 
        # Update threshold list
        self.thresholds = [t for t in thresholds if t[0] in self.df.columns]

        # Remove score transformation in both S-CAP and TrAP,
        # if still present after NA removal
        self.s_cap_pred, self.trap_pred = [], []

        if "S-CAP" in self.df.columns:
            self.s_cap_pred = self.df['S-CAP'].copy().to_numpy()
            self.df.loc[:, 'S-CAP'] = self.df['S-CAP'].apply(lambda x: x - 1 if x > 1 else x)

        if "TraP" in self.df.columns:
            self.trap_pred = self.df['TraP'].copy().to_numpy()
            self.df.loc[:, 'TraP'] = self.df['TraP'].apply(lambda x: x - 1 if x > 1 else x)

        mapper_X, mapper_y = self.build_converter()
        self.X = mapper_X.fit_transform(self.df)
        self.y = mapper_y.fit_transform(self.df).ravel()
        self.features = mapper_X.transformed_names_

        if undersample:
            self.X_resampled, self.y_resampled = self.undersample_data()

        data_to_write = np.append(self.X, self.y.reshape(self.y.shape[0], -1), axis=1)
        pd.DataFrame(data_to_write, columns=self.features + ['label']).to_csv(os.path.join(self.out_dir, 'ml_processed_data_used.tsv'), sep="\t", index=False)       
    
    def build_converter(self):
        """
        Prepares a mapper between Pandas Dataframe and sklearn matrix
        """

        label_encoding = gen_features(columns=self.label_fields,
                                      classes=[LabelBinarizer])

        categorical = gen_features(columns=[[f] for f in self.categorical_fields],
                                   classes=[{'class': SimpleImputer, 'strategy': "most_frequent"},
                                            {'class': OneHotEncoder, 'sparse': False}],
                                   suffix="_cat")

        numeric = gen_features(columns=[[t[0]] for t in self.thresholds],
                               classes=[{'class': SimpleImputer, 'strategy': 'median'}])

        # if any boolean
        boolean = []
        for f in self.df.columns:
            if self.df[f].dtype == bool and f not in self.label_fields:
                self.df[f] = pd.to_numeric(self.df[f])
                boolean.append(([f], None))

        mapper_X = DataFrameMapper(numeric) if self.only_numeric else DataFrameMapper(categorical + boolean + numeric)
        mapper_y = DataFrameMapper(label_encoding)
        return mapper_X, mapper_y

    def undersample_data(self):
        """
        Performs random undersample of the majority class
        """

        rus = RandomUnderSampler(sampling_strategy="majority",
                                 random_state=0)

        X_resampled, y_resampled = rus.fit_sample(self.X, self.y)
        return X_resampled, y_resampled

    def do_classification(self):
        """
        Split input data and train several well known
        classifiers using a gridsearch approach. More
        evaluations are done with test data which is
        also used to compare against individual tools

        :return dict: Dictionary with trained pipelines
            for each of the tested classifiers
        """
        # Available metrics with: sklearn.metrics.SCORERS.keys()
        scoring_metrics = {'AUC': 'roc_auc',
                           'Accuracy': 'accuracy',
                           'F1_score': 'f1'}
        refit_metric = 'F1_score'

        trained_models = {}
        # Split training and test data to have equally
        # balanced classes in both training and test
        sss = StratifiedShuffleSplit(n_splits=1, test_size=0.2 , random_state=0)

        for train_idx, test_idx in sss.split(self.X, self.y):
            train_X, test_X = self.X[train_idx], self.X[test_idx]
            train_y, test_y = self.y[train_idx], self.y[test_idx]

            #############################
            ## Machine Learning class ###
            #############################
            clf = Classifiers(train_X, train_y, scoring_metrics, refit_metric)

            classifier_list = [#'KNN', 
                               'Naive_Bayes',
                               'LDA',
                               'QDA',
                               'Decision_Tree',
                               'Random_Forest',
                               'Ada_Boost',
                               'SVM',
                               'Logistic_Regression']

            # , 'Genetic_Programming']

            out_metrics = {}
            out_best_params = []
            for _c in classifier_list:
                pipeline_out, best_params, cv_score = clf.train(_c)
                out_best_params.append([_c, ';'.join(['{}={}'.format(k,v) for k, v in best_params.items()]) if best_params else None, cv_score])
                y_pred = pipeline_out.predict(test_X)

                # Update dict like in top_tools
                # (repeated metrics because coverage is 1, thus weighted
                # metrics are the same as the original metric value
                out_metrics[_c] = self.make_metrics(test_y, y_pred)
                trained_models[_c] = pipeline_out

            _clf_performance = self._dict_to_df(out_metrics)
            out_file = os.path.join(self.out_dir, "{}_gridSearch_output.tsv".format(self.location))
            pd.DataFrame.from_records(out_best_params, columns=['Classifier', 'Parameters_of_best_estimator', 'Cross_validation_score']).to_csv(out_file,
                                                                                                                                                index=False,
                                                                                                                                                sep='\t')
            
            #####################################
            ### Top tools overall performance ###
            #####################################
            top_tools_loc = self.top_tools[self.location]
            merged = pd.concat([_clf_performance, top_tools_loc], ignore_index=True)

            assert self.rank_metric in _clf_performance.columns, "Ranking metric provided is not available."

            out_file = os.path.join(self.out_dir, "{}_metrics_clf_along_best_tools.pdf".format(self.location))
            plot_metrics(merged, out_file, self.rank_metric)

            ##########################################
            ### Top tools performance on TEST data ###
            ##########################################
            # Train Single tool classifier for fair comparison
            single_tool_classifiers = self.expand_classifiers_list()
            out_metrics = {}
            for _c in single_tool_classifiers:
                _name = _c[0]
                assert _name in self.features, "{} scores must be present in the " \
                                               "training data to predict with single " \
                                               "feature classifier".format(_name)
                _clf = _c[1]
                _clf.fit(train_X, train_y)

                # Recover pathogenic predictions made by multi threshold tools
                # (pathogenic predictions transformed (+1))
                if _name == "TraP":

                    m = np.median(self.trap_pred[~np.isnan(self.trap_pred)])
                    self.trap_pred[np.isnan(self.trap_pred)] = m
                    test_x_single_feature = self.trap_pred[test_idx].reshape(-1, 1)

                elif _name == "S-CAP" or _name == "SCAP":
                    m = np.median(self.s_cap_pred[~np.isnan(self.s_cap_pred)])
                    self.s_cap_pred[np.isnan(self.s_cap_pred)] = m
                    test_x_single_feature = self.s_cap_pred[test_idx].reshape(-1, 1)

                else:
                    i = self.features.index(_name)
                    test_x_single_feature = np.expand_dims(test_X[:, i], axis=1)

                y_pred = _clf.predict(test_x_single_feature)
                out_metrics[_name] = self.make_metrics(test_y, y_pred)

            single_tool_perfor = self._dict_to_df(out_metrics)
            merged = pd.concat([_clf_performance, single_tool_perfor], ignore_index=True)

            out_file = os.path.join(self.out_dir, "{}_metrics_clf_only_on_test_data.pdf".format(self.location))
            merged.to_csv(os.path.join(self.out_dir, "{}_metrics_clf_only_on_test_data_N_variants_equals_to_{}.tsv".format(self.location, test_X.shape[0])), sep="\t", index=False)
            plot_metrics(merged, out_file, self.rank_metric)
            
        return trained_models

    def expand_classifiers_list(self):
        """
        Expand classifiers list considering the top individual
        tools as single classifiers.
        :return:
        """
        extra_classifiers = []
        _df_top = self.top_tools[self.location]
        for tool, direction, threshold, *args in self.thresholds:

            if tool in _df_top.tool.values:
                single_cls = SingleFeatureClassifier(tool, direction, threshold)
                extra_classifiers.append((tool, single_cls))

        return extra_classifiers

    def make_metrics(self, y_true: np.array, y_pred: np.array) -> list:
        """
        Generate common metrics for a given classifier

        :param np.array y_true: True labels
        :param np.array y_pred: Predicted labels

        :return list: List with some metrics
        """

        tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
        specificity = round(utils_tools.ratio(tn, tn + fp), 3)
        sensitivity = round(utils_tools.ratio(tp, tp + fn), 3)
        precision = round(utils_tools.ratio(tp, tp + fp), 3)
        accuracy = round(accuracy_score(y_true, y_pred), 3)
        f1 = round(f1_score(y_true, y_pred), 3)
        coverage = 1
        mcc = round(matthews_corrcoef(y_true, y_pred), 3)
        normalized_mcc = (mcc + 1) / 2
        weighted_norm_mcc = round(normalized_mcc * coverage, 3)

        
        return [accuracy, accuracy, f1, f1, coverage,
                specificity, sensitivity, precision,
                mcc, normalized_mcc, weighted_norm_mcc]

    def _dict_to_df(self, input_d: dict) -> pd.DataFrame:
        """
        Converts a dict of statistics into a df

        :param dict input_d: Input dict
        :return pd.DataFrame: Output df
        """
        return pd.DataFrame.from_dict(input_d,
                                      orient='index',
                                      columns=['weighted_accuracy',
                                               'accuracy',
                                               'weighted_F1',
                                               'F1',
                                               'coverage',
                                               'specificity',
                                               'sensitivity',
                                               'precision',
                                               'mcc',
                                               'normalized_mcc',
                                               'weighted_norm_mcc']).rename_axis('tool').reset_index()

    def generate_feature_ranking(self, trained_clfs: dict):
        """
        Apply methodologies to rank features
        and identify the most important tools for predictions

        :param dict trained_clfs: Dict with trained classifiers
        """
        self.out_dir = os.path.join(self.out_dir, "feature_ranking")
        os.makedirs(self.out_dir, exist_ok=True)
        self.information_gain()

        try:
            rf_pipeline = trained_clfs['Random_Forest']
            self.feature_importance(rf_pipeline)

        except KeyError:
            pass

        self.recursive_feature_elimination(trained_clfs)

    def information_gain(self):
        """
        Calculates mutual information for the target label.
        Uses all the data
        """
        ig = mutual_info_classif(self.X, self.y, discrete_features=False, random_state=0)
        indices = np.argsort(ig)
        sorted_tools = [self.features[i] for i in indices]
        plot_feature_importance(dict(zip(sorted_tools, ig[indices])),
                                'information_gain',
                                self.location,
                                self.out_dir)

    def feature_importance(self, rf_pipeline: Pipeline):
        """
        Calculates feature importance scores based
        on the random forest pipeline and generates
        corresponding plot

        :param rf_pipeline: A trained Random forest pipeline
        :return:
        """
        _clf = rf_pipeline.named_steps['rf']
        indices = np.argsort(_clf.feature_importances_)
        sorted_tools = [self.features[i] for i in indices]
        importance_map = dict(zip(sorted_tools, _clf.feature_importances_[indices]))
        std = np.std([_clf.feature_importances_ for _tree in _clf.estimators_], axis=0)[indices]
        out_dir = os.path.join(self.out_dir, "random_forest")
        os.makedirs(out_dir, exist_ok=True)
        plot_feature_importance(importance_map,
                                'feature_importance',
                                self.location,
                                out_dir,
                                std=std)

    def recursive_feature_elimination(self, trained_clfs: dict):
        """
        Performs recursive feature elimination by fitting some
        of the classifiers provided and iteratively removing one
        or a subset of features.

        Fitting is done with best parameters found during the
        gridsearch for initial classification

        :param dict trained_clfs: Dict with trained classifiers
        """
        for _name, _clf in trained_clfs.items():
            new_clf, selector = "", ""
            if _name == "Random_Forest":
                out_dir = os.path.join(self.out_dir, 'random_forest')
                os.makedirs(out_dir, exist_ok=True)
                best_params = _clf.named_steps['rf'].get_params()
               
                new_clf = RandomForestClassifier(**best_params)
                selector = RFECV(new_clf,
                                 min_features_to_select=1,
                                 cv=StratifiedKFold(),
                                 n_jobs=-1)

                selector = selector.fit(self.X, self.y)

            elif _name == "Logistic_Regression":
                out_dir = os.path.join(self.out_dir, 'logistic_regression')
                os.makedirs(out_dir, exist_ok=True)
                # Couldn't use a pipeline in the RFECV, therefore
                # data needs to be scaled separately here
                scaler = StandardScaler()
                _X = scaler.fit_transform(self.X)
                best_params = _clf.named_steps['lr'].get_params()
        
                new_clf = LogisticRegression(**best_params)

                selector = RFECV(new_clf,
                                 min_features_to_select=1,
                                 cv=StratifiedKFold(),
                                 n_jobs=-1)

                selector = selector.fit(_X, self.y)

            else:
                continue

            plot_rfecv(selector, 1, self.features,
                       _name, self.location, out_dir)

    def generate_tree_plot(self, trained_clfs: dict):
        """
        Generate plot from a fitted decision tree model

        :param dict trained_clfs: Dict with trained classifiers
        """
        plot_decision_tree(trained_clfs['Decision_Tree'],
                           self.X,
                           self.y,
                           self.features,
                           self.location,
                           self.out_dir)


class SingleFeatureClassifier(dummy.DummyClassifier):
    """
    Class to create a Single Feature classifier
    based on the scores of individual tools
    """

    def __init__(self, tool, direction, threshold):
        super(SingleFeatureClassifier, self).__init__()
        self.tool = tool
        self.direction = direction
        self.threshold = float(threshold)

    def meaning(self, x):
        """
        Dummy decision function

        :param float x: Feature value of a single instance
        :return:
        """
        if self.direction == ">":
            return True if x >= self.threshold else False
        else:
            return True if x < self.threshold else False

    def fit(self, X: np.array, y: np.array, **kwargs):
        """
        Fit a dummy estimator without actually training anything

        :param np.array X: Training features
        :param np.array y: Training labels
        :param kwargs: Extra args
        :return:
        """
        X, y = check_X_y(X, y)
        self.classes_ = unique_labels(y)
        self.X_ = X
        self.y_ = y
        return self

    def predict(self, X: np.array) -> np.array:
        """
        Makes prediction over an array with single feature values
        :param np.array X: Input array
        :return np.array: Array with predictions
        """
        check_is_fitted(self)
        X = check_array(X)
        return np.array([self.meaning(x) for x in X]).astype(int)



import os
from collections import defaultdict
from typing import Union, List, Tuple

import hgvs
import pandas as pd
from importlib import resources
import config

from predictions.filters import update_thresholds, subset_toolset_by_scope, \
    subset_variants_by_type, _extract_possible_filters
from preprocessing.clinvar import get_clinvar_cached, remove_clinvar_useless
from preprocessing.location import *
from preprocessing.osutils import check_file_exists, setup_output_directory, ensure_folder_exists
from preprocessing.utils_tools import *
from preprocessing.vcf import process_vcf

TOOLS_CONFIG = resources.open_text(config, 'tools_config.txt')

class Base(object):
    """
    Base class to parse input args and
    create the dataframe for downstream
    analysis
    """

    def __init__(self, vcf: Union[Tuple, str],
                 out_dir: str,
                 scopes_to_predict: List = None,
                 types_of_variant: List = None,
                 metric: str = "weighted_accuracy",
                 location: str = "HGVSc",
                 aggregate_classes: bool = False,
                 select_conseqs: str = "gene_body",
                 do_intronic_analysis: bool = False,
                 split_splice_sites: bool = False,
                 is_clinvar: bool = False,
                 allele_frequency_col: str = "gnomADg_AF",
                 skip_heatmap: bool = False,
                 tools_config: str = TOOLS_CONFIG,
                 interrogate_mode: bool = False):

        """
        :param Union[Tuple, str] vcf: Input VCF(s) to analyse.
        :param str out_dir: Path to the output directory.

        :param List scopes_to_predict: Restrict analysis
            to a subset of tools based on their scope.
            Available options: ['Conservation', 'Whole_genome',
            'Protein', 'Splicing']. Default: `None` tools
            from all scopes are used.

        :param List types_of_variant: Restrict analysis to the
            given variant types. Available options: ['snps',
            'indels', 'insertions', 'deletions', 'mnps'].
            Default: `None`, Performance is measured for all
            types together and for each subtype separately.

        :param str metric: Metric to rank the tools.
            Available options: ['weighted_accuracy', 'accuracy',
            'F1', 'weighted_F1', 'coverage']. Default:
            `weighted_accuracy`.

        :param str location: VEP annotated VCF field to
            extract location of the variants. Available
            options: ['HGVSc, 'Consequence']. Default: "HGVSc".

        :param bool aggregate_classes: Aggregate coding and splice region
            variant classes into more high-level concepts to be analyzed 
            together
        
        :paran str select_conseqs: How to select the top consequence
            block per variant
        
        :param bool do_intronic_analysis: Whether additional analysis of
            intronic variants extracted from HGVSc field will
            be performed
            
        :param bool split_splice_sites: Do separate analysis for donor and 
        acceptor variants when 'do_intronic_analysis' is set
            
        :param bool is_clinvar: Whether `vcf` is from Clinvar.
            Default: `False`

        :param str allele_frequency_col: VCF field that measures
            population frequency of variants. If not present,
            analysis that depend of the field will be ignored.
            Default: `gnomADg_AF`

        :param bool skip_heatmap: Whether drawing of heatmaps
            should be skipped. Default: `False`

        :param str tools_config: Path to the tools config
            where available tools in VCF are mapped to the
            corresponding VCF annotation
            
        :param bool interrogate_mode: Whether veta run
            is in interrogate mode
            
        :return pd.DataFrame: Processed dataframe
        """
        self.metric = metric
        self.location_from = location
        self.do_intronic_analysis = do_intronic_analysis
        self.split_splice_sites = split_splice_sites
        self.aggregate_classes = aggregate_classes
        self.select_conseqs = select_conseqs
        self.location_filters = _extract_possible_filters(self.aggregate_classes)

        self.is_clinvar = is_clinvar
        self.is_interrogate_mode = interrogate_mode
        self.out_dir = setup_output_directory(out_dir)
        ensure_folder_exists(self.out_dir)
        self.allele_frequency_col = allele_frequency_col
        self.skip_heatmap = skip_heatmap

        tools_config = self._parse_tools_config(tools_config)
        self.thresholds = update_thresholds(tools_config)
        self.thresholds = subset_toolset_by_scope(self.thresholds, scopes_to_predict)
        self.available_tools = [t[0] for t in self.thresholds]

        # TODO change global thresholds list to remove excess of fields when custom models are provided
        self.tools_config = {k: v for k, v in tools_config.items() if k in self.available_tools}
        self.variant_types = subset_variants_by_type(types_of_variant)

        if self.location_from == "Consequence" and self.do_intronic_analysis:
            raise ValueError('If \'--do_intronic_analysis\' is set, \'--location\' '
                             'must be \'HGVSc\' because intronic bin are assigned '
                             'based on the distance offset extracted from HGVSc expressions.')

        if isinstance(vcf, str):
            check_file_exists(vcf)
            # if it's clinvar, additional
            # processing will be done
            self.df = self.get_df_ready(vcf)

        else:
            _dfs = []
            for file in vcf:
                f = os.path.basename(file).lower()
                is_benign = False if any(x in f for x in ["pathogenic", "deleterious"]) else True
                _dfs.append(self.get_df_ready(file, is_benign=is_benign))
        
            self.df = pd.concat(_dfs).reset_index(drop=True)

        # Replace missing AF with 0 for AF plots
        if self.allele_frequency_col in self.df.columns:

            if self.df[self.allele_frequency_col].dtype == object:
                self.df[self.allele_frequency_col] = self.df[self.allele_frequency_col].\
                    apply(lambda x: [0 if v is None else x[0] for v in x][0])

        # Replace missing AF if AF is used as a tool itself to evaluate performance
        if self.allele_frequency_col in [v[0][0] for v in self.tools_config.values()]:
            name = [tool for tool, v in self.tools_config.items() if v[0][0] == self.allele_frequency_col][0]
            # If tool name is different than vcf field (it it is the same, it's dealt above)
            try:
                if name != self.allele_frequency_col:
                    self.df[name] = self.df[name].fillna(0)
            except KeyError:
                raise KeyError('Unknown column {} in the processed dataframe. '
                               'Remove *tsv file and run again Clinvar analysis '
                               'with the added tool.'.format(name))

        if self.is_clinvar:
            # cache processed df
            self.df.to_csv(vcf + '.tsv', index=False)

    def get_df_ready(self, vcf: str, is_benign: bool = False):
        """
        Process input VCF(s) so that df are ready for
        downstream analysis

        :param str vcf: Input VCF file
        :param bool is_benign: Whether `vcf` refers to a set of
            benign variants. Only useful when `vcf` is not from
            Clinvar. Default: `False`
        :return pd.DataFrame: Cleaned dataframe
        """
        logging.info("Processing {} file".format(os.path.basename(vcf)))

        if self.is_clinvar:
            _df = get_clinvar_cached(vcf)
            # If cached file was found, DF is ready
            if _df is not None:
                return _df

        df, fields_missing_in_vcf = process_vcf(vcf, 
                                                tools_config=self.tools_config,
                                                thresholds=self.thresholds,
                                                select_conseq = self.select_conseqs,
                                                is_clinvar=self.is_clinvar,
                                                af_col=self.allele_frequency_col)

        # Remove tool from config and thresholds if corresponding field is absent from at least one of the VCFs
        self.tools_config = {k: v for k, v in self.tools_config.items() if k not in fields_missing_in_vcf}
        self.thresholds = [x for x in self.thresholds if x[0] not in fields_missing_in_vcf]

        if self.is_clinvar:
            df = df.dropna(subset=["CLNSIG", "CLNREVSTAT"])
            df = remove_clinvar_useless(df)
            logging.info("Number of variants after removing "
                         "irrelevant CLNSIG values: {}".format(df.shape[0]))

        # Extract variants
        logging.info("Assigning variants location/type")
        if self.location_from == "HGVSc":
            hp = hgvs.parser.Parser()

            df['location'] = df.apply(get_location,
                                      hp=hp,
                                      aggregate=self.aggregate_classes,
                                      axis=1,
                                      result_type='expand')
            
            if self.do_intronic_analysis:
                df[['intron_bin', 'intron_offset', 'which_ss']] = df.apply(
                    lambda x: assign_intronic_bins(x['HGVSc'], 
                                                   hp, 
                                                   x['location'], 
                                                   self.aggregate_classes), axis=1)

        elif self.location_from == "Consequence":
            df['location'] = df['Consequence'].apply(get_location_from_consequence, 
                                                     detailed=self.detailed_conseqs)

        # Clean predictions scores
        df = self._clean_scores(df.reset_index())

        # Set labels
        if self.is_clinvar:
            df['label'] = (df['CLNSIG'].str.contains('Pathogenic')) | \
                          (df['CLNSIG'].str.contains('Likely_pathogenic'))

        else:
            df['label'] = False if is_benign else True

        booleanDictionary = {True: 'Pathogenic', False: 'Benign'}
        df["outcome"] = df["label"].map(booleanDictionary)
    
        return df

    def _clean_scores(self, df: pd.DataFrame):
        """
        Transform tools scores so that proper
        evaluations can be performed.

        By default, VETA has some methods
        ready to process the scores of specific
        tools.

        :param pd.DataFrame df: Input dataframe
        with information about each variant.

        :return pd.DataFrame: Processed df.
        """

        # pd.options.display.float_format = '{:,.6f}'.format

        # values: [function_name, get_max, get_absolute_value]
        available_functions = {'to_numeric': [to_numeric, None, None],
                               'top_max': [get_top_pred, True, True],
                               'top_min': [get_top_pred, False, False],
                               'top_min_abs': [get_top_pred, False, True],
                               'categorical_to_numeric': [categorical_to_numerical, None, None],
                               'carol_like': [process_condel_carol, None, None],
                               'kipoi_like': [process_kipoi_tools, None, None],
                               'spliceai': [process_spliceai, None, None],
                               'trap': [process_trap, None, None],
                               'scap': [process_scap, None, None],
                               'conspliceml': [process_conspliceml, None, None],
                               'conspliceml_like': [process_conspliceml_like, None, None],
                               'pangolin': [process_pangolin, None, None],
                               'spip': [process_spip, None, None],
                               'labranchor': [process_labranchor, None, None],
                               'absplice_dna': [process_absplice, None, None]
                               }

        logging.info("Engineering the scores.")
        clean_functions = {
            "GERP": available_functions['top_max'],
            "phyloP": available_functions['top_max'],
            "phastCons": available_functions['top_max'],
            "SiPhy": available_functions['to_numeric'],
            "CDTS": available_functions['to_numeric'],

            "LRT": available_functions['to_numeric'],
            "Sift": available_functions['top_min_abs'],
            "Polyphen2HVAR": available_functions['top_max'],
            "Polyphen2HDIV": available_functions['top_max'],
            "MutationAssessor": available_functions['top_max'],
            "MutationTaster2": available_functions['top_max'],
            "FATHMM": available_functions['top_min'],
            "Provean": available_functions['top_min'],
            "Mutpred": available_functions['to_numeric'],
            "CAROL": available_functions['carol_like'],
            "Condel": available_functions['carol_like'],
            "M-CAP": available_functions['to_numeric'],
            "MetaLR": available_functions['to_numeric'],
            "MetaSVM": available_functions['to_numeric'],
            "REVEL": available_functions['top_max'],
            "VEST4": available_functions['to_numeric'],
            "CardioBoost": available_functions['to_numeric'],
            "CardioVAI": available_functions['to_numeric'],
            "PrimateAI": available_functions['to_numeric'],
            "VARITY": available_functions['to_numeric'],
            "MutFormer": available_functions['to_numeric'],
            "MutScore": available_functions['to_numeric'],
            "MVP": available_functions['to_numeric'],
            "MTR": available_functions['to_numeric'],
            "MPC": available_functions['to_numeric'],
            "MISTIC": available_functions['top_max'],
            "ClinPred": available_functions['top_max'],
            "cVEP": available_functions['categorical_to_numeric'],
            "EVE_Class20": available_functions['categorical_to_numeric'],
            "EVE_Class50": available_functions['categorical_to_numeric'],
            "EVE_Class90": available_functions['categorical_to_numeric'],
            "EVE": available_functions['to_numeric'],
            
            "fitCons": available_functions['to_numeric'],
            "LINSIGHT": available_functions['to_numeric'],
            "GWAVA": available_functions['to_numeric'],
            "CADD_v1.5": available_functions['to_numeric'],
            "CADD-Splice": available_functions['to_numeric'],
            "Eigen": available_functions['to_numeric'],
            "FATHMM-MKL": available_functions['to_numeric'],
            "FunSeq2": available_functions['top_max'],
            "DANN": available_functions['to_numeric'],
            "ReMM": available_functions['to_numeric'],
            "CAPICE": available_functions['top_max'],

            "HAL": available_functions['kipoi_like'],
            "S-CAP": available_functions['scap'],
            "MMSplice": available_functions['kipoi_like'],
            "kipoiSplice4": available_functions['kipoi_like'],
            "kipoiSplice4_cons": available_functions['kipoi_like'],
            "TraP": available_functions['trap'],
            "SPIDEX": available_functions['to_numeric'],
            "dbscSNV": available_functions['top_max'],
            "MaxEntScan": available_functions['top_max'],
            "SpliceAI": available_functions['spliceai'],
            "SQUIRLS": available_functions['to_numeric'],
            "ConSpliceML": available_functions['conspliceml'],
            "IntSplice2": available_functions['to_numeric'],
            "CI-SpliceAI": available_functions['spliceai'],
            "Pangolin": available_functions['pangolin'],
            "SPiP": available_functions['spip'],
            "MLCsplice": available_functions['top_max'],
            "AbSplice-DNA": available_functions['absplice_dna'],
            "PDIVAS": available_functions['conspliceml_like'],
            "LaBranchoR": available_functions['labranchor']
        }

        _functions_that_require_loc = ["process_trap"]
        _functions_that_require_symbol = ["process_spliceai", "process_conspliceml"]
        _absent_tools = {_tool: info for _tool, info in self.tools_config.items()
                         if _tool not in clean_functions.keys()}

        # If new custom tool
        if _absent_tools:
 
            # get function, if provided (5th col in config)
            for _tool, info in _absent_tools.items():
                if len(info) > 4:    
                    _function = info[4]
                    #self.tools_config[_tool] = info[:-1]
                else:
                    logging.info('Processing function was not provided for {}. Using the default to_numeric function. Be careful.'.format(_tool))
                    _function = 'to_numeric'
                clean_functions[_tool] = available_functions[_function]
        # For each tool belonging to the given scope
        for _tool, info in self.tools_config.items():
            # if it is known tool
            if _tool in clean_functions.keys():
                # get function to apply
                _f = clean_functions[_tool][0]
                # is_max
                is_max = clean_functions[_tool][1]
                # is_absolute
                is_absolute = clean_functions[_tool][2]

                # if the following tools, use custom methods
                # that need variant location
                if _f.__name__ in _functions_that_require_loc:
                    df[_tool] = df[[_tool] + ['location']].apply(_f, axis=1)

                elif _f.__name__ in _functions_that_require_symbol:
                    
                    df[_tool] = df[[_tool] + ['SYMBOL']].apply(_f, axis=1)
 
                # apply single function to
                # the target columns
                else:
                    # if max and absolute info are not
                    # required (e.g. to_numeric, and custom
                    # function to specific tools)
                    if all([_v is None for _v in [is_max, is_absolute]]):
                        df[_tool] = df[_tool].apply(_f)
                                   
                    else:
                        df[_tool] = df[_tool].apply(_f, is_max=is_max,
                                                    absolute=is_absolute)

        return df

    def _parse_tools_config(self, config):
        """
        Parse tools config file to ensure
        correctness for downstream analysis.
        - Empty and commented lines (^#) are ignored.
        - First two fields are required
        - If tool name (1st col) not in the list
        of available tools, it expects to be a new
        custom method so that, 3rd (scope) and 4th
        and 5th columns are required.

        :param str config: Input file
        :return defaultdict: Dict with
        info about each tool
        """
        AVAILABLE_TOOLS_NAMES = ["GERP", "phyloP", "phastCons", "SiPhy",
                                 "CDTS", "LRT", "Sift", "Polyphen2HVAR",
                                 "Polyphen2HDIV", "MutationTaster2",
                                 "MutationAssessor", "FATHMM", "Provean",
                                 "Mutpred", "CAROL", "Condel", "M-CAP",
                                 "MetaSVM", "MetaLR", "REVEL", "VEST4",
                                 "MVP", "PrimateAI", "cVEP", "VARITY",
                                 "MTR", "MPC", "MutScore", "MutFormer",
                                 "ClinPred", "MISTIC",
                                 "EVE", "EVE_class20", "EVE_class50","EVE_class90",                             
                                 "CardioBoost", "CardioVAI",
                                 "fitCons", "LINSIGHT", "ReMM", "GWAVA", "FATHMM-MKL",
                                 "Eigen", "FunSeq2", "CADD_v1.5", "CADD-Splice", "DANN", "CAPICE",
                                 "HAL", "SPIDEX", "dbscSNV", "MaxEntScan", "SpliceAI", "S-CAP", "ConSpliceML",
                                 "TraP", "MMSplice", "SQUIRLS", "IntSplice2", "kipoiSplice4", "kipoiSplice4_cons",
                                 "CI-SpliceAI", "Pangolin", "SPiP", "LaBranchoR", "MLCsplice", "PDIVAS", "AbSplice-DNA"]

        AVAILABLE_SCOPES = ["Protein", "Conservation", "Whole_genome", "Splicing"]
        AVAILABLE_FUNCTIONS = ['to_numeric', 'top_max', 'top_min', 'top_min_abs',
                               'categorical_to_numeric', 'kipoi_like', 'carol_like', 
                               'trap', 'scap', 'conspliceml', 'conspliceml_like',
                               'spliceai', 'ci_spliceai',
                               'pangolin', 'spip', 'labranchor']

        data = defaultdict(list)
        if isinstance(config, str):
            _c = open(config, 'r')
            config = _c

        for line in config:
            line = line.rstrip()
            if line.startswith("#") or not line:
                continue

            _fields = line.split("\t")

            assert len(_fields) >= 2, "First 2 fields are required in the tools config file. " \
                                      "Problematic line: {}.\nMake sure fields are tab separated.".format(line)

            assert _fields[0] not in data.keys(), "Repeated tool name found in the tools " \
                                                  "config file: {}".format(_fields[0])

            data[_fields[0]].append(_fields[1].split(","))
            if _fields[0] not in AVAILABLE_TOOLS_NAMES:
                assert len(_fields) >= 5, "\'{}\' tool is not in the list of available " \
                                          "tools. Thus, It is assumed to be a custom method " \
                                          "provided by the user, which means that a reference " \
                                          "threshold, its directionality and scope must be provided. " \
                                          "Please set that info (at least 5 columns), or check if the " \
                                          "tool name was just misspelled. List of default tools: " \
                                          "{}".format(_fields[0], '\n' + '\n'.join(AVAILABLE_TOOLS_NAMES))

            if len(_fields) > 2:
                assert _fields[2] in ['>', "<"], "Directionality of the threshold (3th col) in " \
                                                 "the tools config must be set to '>' or '<' at " \
                                                 "line:\n{}".format(line)

                try:
                    _ref_thresh = float(_fields[3])
                except ValueError:
                    raise ValueError("Reference threshold (4th col) in the tools config "
                                     "file must be a number at line:\n{}".format(line))

                data[_fields[0]].extend([_fields[2], _ref_thresh])

                if len(_fields) > 4:
                    assert _fields[4] in AVAILABLE_SCOPES, "\'{}\' is not in the list of " \
                                                           "available scopes: {} at line:\n{}".format(_fields[4],
                                                                                                      AVAILABLE_SCOPES,
                                                                                                      line)
                    data[_fields[0]].append(_fields[4])

                if len(_fields) > 5:
                    assert _fields[5] in AVAILABLE_FUNCTIONS, "\'{}\' is not in the list of " \
                                                              "available functions: {} at " \
                                                              "line:\n{}".format(_fields[5], AVAILABLE_FUNCTIONS,
                                                                                 line)
                    data[_fields[0]].append(_fields[5])
    
        if len(data) < 2 and self.is_interrogate_mode is False:
            raise ValueError('Config file requires at least two tools to be compared.')

        return data




import logging
import os
import tempfile
from itertools import zip_longest
from typing import Union


def check_file_exists(filename: str = None):
    """Checks if file exists in path"""
    if filename is None:
        return

    if not os.path.isfile(filename):
        raise ValueError("{} file does not exist.".format(filename))
    return filename


def ensure_folder_exists(filename):
    if os.path.isdir(filename):
        pass
    elif os.path.isfile(filename):
        ensure_folder_exists(os.path.dirname(filename))
    else:
        head, tail = os.path.split(filename)
        if head and not os.path.isdir(head):
            ensure_folder_exists(head)
        if tail:
            os.mkdir(filename)


def setup_output_directory(directory: Union[str, None]):
    """
    Creates output directory from given
    name.
    :param Union[str, None] directory: Name of
        the output dir. If `None`, 'out_VETA'
        will be created.
    """
    out_dir = directory if directory is not None else os.path.join(os.getcwd(), "out_VETA")
    # assert not os.path.isdir(out_dir), "Output directory {} exists. " \
    #                                    "Remove it or set a new one.".format(out_dir)
    return out_dir


def split_file_in_chunks(file: str,
                         header: list,
                         n_variants: int,
                         chunk_size: int = 2000):
    """
    Receives
    :param str file: Input VCF to split without the header
    :param list header: VCF header as an iterable
    :param int n_variants: Number of variants in the input `file`
    :param int chunk_size: Number of variants per each new smaller file. Default: `2000`

    :return list: List with tmp file names created
    """
    # Output Files
    outlist = []

    # Ceil division
    n_chunks = -(-n_variants // chunk_size)
    logging.info("More than 5000 variants found ({}). "
                 "Spliting file in {} chunks of {} variants".format(n_variants, n_chunks, chunk_size))

    # Decode header
    _header = [x.decode('utf-8') for x in header]

    # Process variants
    with open(file) as f:
        for i, g in enumerate(grouper(n_chunks, f, fillvalue=''), 1):
            with tempfile.NamedTemporaryFile('w', delete=False) as tmp:
                tmp.writelines(_header)
                tmp.writelines(g)
            outlist.append(tmp.name)

    return outlist


def grouper(n, file_iterable, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    args = [iter(file_iterable)] * n
    return zip_longest(fillvalue=fillvalue, *args)


def print_clinvar_levels():
    """
    Prints the clinvar levels available
    to filter the input variants for all
    the analysis when the reference dataset
    is the clinvar database.
    """
    print("0s ---- Whole clinvar with Pathogenic and Benign assignments" + "\n" +
          "1s --- clinvar with 1 star or more" + "\n" +
          "2s --- 2 stars or more" + "\n" +
          "3s --- 3 stars or more" + "\n" +
          "4s --- 4 stars" + "\n" +
          "0s_l ---- 0 star or more with likely assignments" + "\n" +
          "1s_l ---- 1 star or more with likely assignments" + "\n" +
          "2s_l ---- 2 stars or more with likely assignments" + "\n" +
          "3s_l ---- 3 stars or more with likely assignments" + "\n" +
          "4s_l ---- 4 stars with likely assignments")
    exit(1)



import logging
import pandas as pd
import multiprocessing
from functools import partial
from collections import defaultdict
from typing import List
from preprocessing.osutils import ensure_folder_exists
from plots.plots_intronic_analysis import *
from plots.plots_benchmark_mode import (
    plot_allele_frequency,
    plot_unscored,
    plot_metrics,
    plot_curve,
)
from predictions import metrics
from predictions.apply import apply_tool_predictions


def _parallel_roc(thresholds: list, _df_i_bin: pd.DataFrame, **kwargs):

    roc_metrics, pr_metrics = [], []
    statistics = defaultdict(list)
    na = {}
    tool, direction = thresholds[0], thresholds[1]

    try:
        _no_null_df = _df_i_bin.loc[
            pd.notnull(_df_i_bin[tool + "_prediction"]),
        ].copy()
    except KeyError:
        na[tool] = 1

    statistics = metrics.generate_statistics(
        _df_i_bin, statistics, kwargs["_bin"], tool
    )
    stats_df = pd.DataFrame(statistics)

    na[tool] = stats_df.loc[stats_df["tool"] == tool, "fraction_nan"].iloc[0]
    df_tool = _no_null_df[~_no_null_df[tool + "_prediction"].isnull()]

    ####################
    ### ROC analysis ###
    ####################
    if kwargs["skip_roc"] is False:
        n_pos_pred = df_tool[df_tool[tool + "_prediction"]].shape[0]
        n_neg_pred = df_tool[df_tool[tool + "_prediction"] == False].shape[0]

        if float(na[tool]) > 0.5:
            logging.info(
                "ROC analysis will be skipped for {} at {} bin. "
                "More than 50% of missing data ({})".format(
                    tool, kwargs["_bin"], round(float(na[tool]), 3)
                )
            )

        elif min(n_pos_pred, n_neg_pred) < 10:
            logging.info(
                "ROC analysis will be skipped for {} at {} bin. "
                "No minimum number of predictions on each class (10) found (N={})".format(
                    tool, kwargs["_bin"], min(n_pos_pred, n_neg_pred)
                )
            )

        else:
            try:
                roc_curve, pr_curve, roc_auc, ap_score = metrics.do_roc_analysis(
                    df_tool[[tool, "label"]], tool, higher_is_better=direction == ">"
                )
                roc_metrics = [
                    tool,
                    float(na[tool]),
                    roc_curve[0],
                    roc_curve[1],
                    roc_curve[2],
                    roc_curve[3],
                    roc_auc,
                ]
                pr_metrics = [
                    tool,
                    float(na[tool]),
                    pr_curve[0],
                    pr_curve[1],
                    pr_curve[2],
                    pr_curve[3],
                    ap_score,
                ]

            # E.g. S-CAP
            except TypeError:
                pass

    f1_score = stats_df.loc[stats_df["tool"] == tool, "F1"].iloc[0]
    weighted_f1_score = stats_df.loc[stats_df["tool"] == tool, "weighted_F1"].iloc[0]

    return (
        stats_df,
        roc_metrics,
        pr_metrics,
        [tool, f1_score, weighted_f1_score, na[tool]],
    )


class IntronicAnalysis(object):
    """
    Base class for intronic analysis
    """

    def __init__(
        self,
        df: pd.DataFrame,
        split_splice_sites: bool,
        thresholds: List,
        metric: str,
        aggregate_classes: str,
        out_dir: str,
        af_column: str,
    ):
        """
        Perform analysis of intronic variants
        in a bin-based manner

        :param pd.DataFrame df: Input df
        :param bool split_splice_sites: 
        :param List thresholds: List with
            the reference thresholds
        :param bool aggregate_classes:
        :param str metric: Metric to rank the
            tools in the metrics plot
        :param str out_dir: Output directory
        :param str af_column: Allele frequency column
        :return:
        """
        self.split_splice_sites = split_splice_sites
        self.thresholds = thresholds
        self.metric = metric
        self.aggregate_classes = aggregate_classes
        self.out_dir = os.path.join(out_dir, "intron_analysis")
        self.af_column = af_column
        self.metrics_per_bin = defaultdict(list)
        
        assert "intron_bin" in df.columns, (
            "Intronic bins are not present in the data."
            "Probably a first run of Clinvar "
            "was performed without --do_intronic_analysis "
            "args. To fix, just remove the 'tsv' in "
            "the input directory and try again."
        )
        # intronic variants
        self.df = df[~df["intron_bin"].isnull()].copy(deep=True)
        self.df = apply_tool_predictions(self.df, self.thresholds)
        logging.info("-------------------------")
        logging.info("Intronic analysis started")
        logging.info("-------------------------")

        loc_in_protein_coding = [
            "splice_site",
            "splice_region",
            "intronic",
            "deep_intronic",
        ]
        pc_is_done = False
        self.bin_to_exclude = ["1-2", "3-10"] if self.aggregate_classes else ["1-10"]
        locations = sorted(self.df.location.unique().tolist())
        locations.insert(0, "all")

        for _class in locations:
            self._class = _class
            ss = None
            if self._class == "all":
                _df_loc = self.df.copy(deep=True)

                if self.split_splice_sites:
                    out_per_bin_metrics = []
                    for ss in ["donor", "acceptor"]:
                        _df = _df_loc[_df_loc.which_ss == ss].copy()
                        out_per_bin_metrics.append(self.per_bin_computations(_df, specific_ss=ss))

                    plot_metrics_by_bin_split_ss(pd.concat(out_per_bin_metrics),
                                                  os.path.join(self.out_dir, "all_bin_together/per_bin"),
                                                  self.aggregate_classes)
        
                    self.donor_vs_acceptor(
                        _df_loc, os.path.join(self.out_dir, "performance_at_fixed_thresh")
                    )
                    ss=None
                self.per_bin_computations(_df_loc, specific_ss=ss)
                    
            # elif self._class in loc_in_protein_coding:
            #     if pc_is_done is False:
            #         _class = "no_UTRs"
            #         _df_loc = self.df[self.df.location.isin(loc_in_protein_coding)].copy(
            #             deep=True
            #         )
            #         pc_is_done = True
            #     else:
            #         continue
            #     self.per_bin_computations(_df_loc, specific_ss=ss)
            # else:
            #      _df_loc = self.df[self.df.location == _class].copy(deep=True)  
            #     self.per_bin_computations(_df_loc, specific_ss=ss)

    def donor_vs_acceptor(self, df: pd.DataFrame, out_dir: str):

        stats_d, stats_a = defaultdict(list), defaultdict(list)
        out_file = os.path.join(out_dir, "donor_vs_acceptor_{}.pdf".format(self._class))

        do_donor, do_acceptor = True, True
        donor = df[df.which_ss == "donor"]
        acceptor = df[df.which_ss == "acceptor"]

        if all(x.shape[0] < 20 for x in [donor, acceptor]):
            logging.info(
                "Not enough intronic variants (N < 30) assigned as donor or acceptor-associated (< 1000bp from splice site). Skipping."
            )
            return

        elif donor.shape[0] < 20:
            do_donor = False
            logging.info(
                "Not enough donor-associated variants (N = {}). Only acceptor-related variants will be analyzed.".format(
                    donor.shape[0]
                )
            )

        elif acceptor.shape[0] < 20:
            do_acceptor = False
            logging.info(
                "Not enough donor-associated variants (N = {}). Only acceptor-related variants will be analyzed.".format(
                    acceptor.shape[0]
                )
            )

        for tool, _, _, _ in self.thresholds:
            if do_donor:
                stats_d = metrics.generate_statistics(donor, stats_d, "all", tool)

            if do_acceptor:
                stats_a = metrics.generate_statistics(acceptor, stats_a, "all", tool)

        stats_donor = pd.DataFrame(stats_d)
        stats_acceptor = pd.DataFrame(stats_a)

        if all(x is True for x in [do_donor, do_acceptor]):
            n_donors = stats_donor["total"].max()
            n_acceptors = stats_acceptor["total"].max()
            title = "N donors = {}; N acceptors = {}".format(n_donors, n_acceptors)

            perf = pd.merge(
                stats_donor[["tool", self.metric]],
                stats_acceptor[["tool", self.metric]],
                on="tool",
                how="left",
            )
            perf.columns = ["Tool", "Donor", "Acceptor"]
            perf = perf.melt(
                id_vars="Tool",
                var_name="Splice site",
                value_name=self.metric.replace("_", " "),
            )

        elif do_donor:
            n_donors = stats_donor["total"].max()
            title = "N donors = {}".format(n_donors)

            perf = stats_donor[["tool", self.metric]]
            perf.columns = ["Tool", "Donor"]
            perf = perf.melt(
                id_vars="Tool",
                var_name="Splice site",
                value_name=self.metric.replace("_", " "),
            )

        elif do_acceptor:

            n_acceptors = stats_acceptor["total"].max()
            title = "N acceptors = {}".format(n_acceptors)

            perf = stats_acceptor[["tool", self.metric]]
            perf.columns = ["Tool", "Donor"]
            perf = perf.melt(
                id_vars="Tool",
                var_name="Splice site",
                value_name=self.metric.replace("_", " "),
            )

        plot_donor_vs_acceptor(perf, self.metric.replace("_", " "), out_file, title)

    def per_bin_computations(self, 
                             _df_loc, 
                             specific_ss=None, 
                             min_variants=20):
        """
        :param int min_variants: Minimum number
        of variants required to do ROC analysis.
        Default: `20`
        """
        out_tsv = os.path.join(self.out_dir, "results_tsv")
        out_af = os.path.join(self.out_dir, "allele_frequency")
        out_fixed_thresh = os.path.join(self.out_dir, "performance_at_fixed_thresh")
        out_roc = os.path.join(self.out_dir, "roc_analysis")
        out_all_bin_agg = os.path.join(self.out_dir, "all_bin_together")

        [
            ensure_folder_exists(path)
            for path in [
                self.out_dir,
                out_tsv,
                out_af,
                out_fixed_thresh,
                out_roc,
                out_all_bin_agg,
            ]
        ]

        _aux_class = (
            self._class
            if specific_ss is None
            else self._class + "_" + specific_ss + "_related"
        )
        print("\n")
        logging.info("--------------")
        logging.info("Looking at {} intronic variants".format(_aux_class))
        logging.info("--------------")
        print()

        if _df_loc.empty:
            logging.info("No {} variants found.".format(_aux_class))
            return

        _df_loc["count_class"] = _df_loc.groupby("outcome")["outcome"].transform("size")

        try:
            plot_general_bin_info(
                _df_loc,
                self.out_dir,
                _aux_class,
                self.aggregate_classes,
                self.af_column,
            )
        except ValueError:
            logging.info("Problem plotting info about intronic bins. Skipping.")

        for _bin, _filter_func in filter_intronic_bins:
            #if "all_except" in _bin:
            #    continue

            if _bin in self.bin_to_exclude:
                continue
            skip_roc = False
            print()
            logging.info("-------")
            logging.info("BIN: {}.".format(_bin))
            logging.info("-------")
            _df_i_bin = _filter_func(_df_loc).copy(deep=True)

            if _df_i_bin.empty:
                logging.info("No {} variants at {} bin.".format(_aux_class, _bin))
                continue

            n_pos = _df_i_bin["label"].sum()
            n_neg = (~_df_i_bin["label"]).sum()
            logging.info("Pathogenic: {}".format(n_pos))
            logging.info("Benign: {}".format(n_neg))
                
            if _df_i_bin.shape[0] < min_variants:
                logging.info(
                    "Not enough {} variants for ROC analysis "
                    "at {} bin (N={}, required={})".format(
                        _aux_class, _bin, _df_i_bin.shape[0], min_variants
                    )
                )
                skip_roc = True

            elif min(n_pos, n_neg) < 10:
                logging.info(
                    "Number of variants in the minority class is lower (N={}) "
                    "than minimum required (N=10) for ROC analysis.".format(
                        min(n_pos, n_neg)
                    )
                )
                skip_roc = True

            kwargs = {
                "_bin": _bin,
                "skip_roc": skip_roc,
            }

            with multiprocessing.Pool() as p:

                stats_df, roc_metrics, pr_metrics, _per_bin = zip(
                    *p.map(
                        partial(_parallel_roc, _df_i_bin=_df_i_bin, **kwargs),
                        self.thresholds,
                    )
                )
                stats_df = pd.concat(stats_df)

            roc_m = plot_curve(
                roc_metrics,
                os.path.join(out_roc, "{}_{}_ROC.pdf".format(_aux_class, _bin)),
                (n_pos, n_neg),
            )

            pr_m = plot_curve(
                pr_metrics,
                os.path.join(out_roc, "{}_{}_ROC_pr.pdf".format(_aux_class, _bin)),
                (n_pos, n_neg),
                is_roc=False,
            )

            stats_out = os.path.join(out_tsv, "statistics_{}_{}.tsv".format(_aux_class, _bin))
            af_plot = os.path.join(out_af, "AF_{}_{}.pdf".format(_aux_class, _bin))
            unscored_plot = os.path.join(out_fixed_thresh, "unscored_fraction_{}_{}.pdf".format(_aux_class, _bin))
            metrics_plot = os.path.join(out_fixed_thresh, "tools_metrics_{}_{}.pdf".format(_aux_class, _bin))

            plot_allele_frequency(_df_i_bin, af_plot, self.af_column)
            plot_unscored(stats_df, unscored_plot)
            plot_metrics(stats_df, metrics_plot, self.metric)
            
            if not _bin in ["all_intronic", "all_except_1-2", "all_except_1-10", None]:

                for tool_info in _per_bin:
                    res = [_aux_class, _bin, tool_info[1], tool_info[2], tool_info[3]]

                    if skip_roc:
                        res.extend([None, None])
                    else:
                        res.append(roc_m[tool_info[0]]) if tool_info[0] in roc_m.keys() else res.append(None)
                        res.append(pr_m[tool_info[0]]) if tool_info[0] in pr_m.keys() else res.append(None)
                        
                    self.metrics_per_bin[tool_info[0]].append(res)
            
            if roc_m:
                stats_df = pd.merge(stats_df, pd.DataFrame.from_dict(roc_m, orient='index', columns=['auROC']), how = 'left', left_on = 'tool', right_index = True)
            else:
                stats_df['auROC'] = None
            
            if pr_m:
                stats_df = pd.merge(stats_df, pd.DataFrame.from_dict(pr_m, orient='index', columns=['auPRC']), how = 'left', left_on = 'tool', right_index = True)
            else:
                stats_df['auPRC'] = None

            stats_df.drop(["filter"], axis=1).to_csv(stats_out, sep="\t", index=False)
            
        df_metrics_per_bin = pd.DataFrame(
            [[k] + i for k, v in self.metrics_per_bin.items() for i in v],
            columns=[
                "tool",
                "variant_class",
                "bin",
                "F1",
                "weighted_F1",
                "fraction_nan",
                "auROC",
                "auPRC"
            ],
        )

        df_metrics_per_bin.groupby("variant_class").apply(
            plot_metrics_by_bin,
            os.path.join(out_all_bin_agg, "per_bin"),
            self.aggregate_classes,
        )
        
        if specific_ss:
            _df = df_metrics_per_bin[df_metrics_per_bin.variant_class.str.contains('related')]
            return _df



import argparse
import sys
import logging
from importlib import resources
import config
from interrogate import PredictionsEval
from benchmark import BenchmarkTools
from preprocessing.osutils import print_clinvar_levels
import importlib.metadata

__version__ = importlib.metadata.version("veta")
hgvs_logger = logging.getLogger('hgvs')
hgvs_logger.setLevel(logging.CRITICAL)
scikit_logger = logging.getLogger('sklearn')
scikit_logger.setLevel(logging.CRITICAL)
logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(asctime)s %(message)s')

CONFIG_PATH = resources.open_text(config, 'tools_config.txt')

def main():
    """
    Main function
    """
    parser = argparse.ArgumentParser(prog="veta",
                                     description='Simple tool to evaluate variant '
                                                 'prediction methods')

    subparsers = parser.add_subparsers(dest="command")

    # Parent subparser. Note `add_help=False` and creation via `argparse.`
    parent_parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__))
    parent_parser.add_argument('-o', '--out_dir', metavar='', help='Path to store all the output results. '
                                                                   'Default: "out_VETA"')

    parent_parser.add_argument('-s', '--scopes_to_evaluate', metavar='',
                               help='Restrict analysis to a subset of tools based on their '
                               'scope. Available options: {%(choices)s}. Default: Tools from '
                               'all scores are used.',
                               nargs='+', choices=('Conservation', 'Whole_genome', 'Protein', 'Splicing'))

    parent_parser.add_argument('-t', '--types_of_variant', metavar='',
                               help='Restrict analysis to the given variant types. '
                               'Available options: {%(choices)s}. Default: Performance '
                               'is measured for all variants and each subtype.',
                               nargs='+', choices=('all_types', 'snps', 'indels', 'insertions', 'deletions', 'mnps'))

    parent_parser.add_argument('-m', '--metric', metavar='',
                               help='Metric to rank the tools. Available options: '
                               '{%(choices)s}. Default: "weighted_accuracy."',
                               choices=('weighted_accuracy', 'accuracy',
                                        'F1', 'weighted_F1', 'coverage',
                                         'norm_mcc', 'weighted_norm_mcc'),

                               default='weighted_accuracy')

    parent_parser.add_argument('-l', '--location', metavar='', default="HGVSc", choices=("HGVSc", "Consequence"),
                               help='VCF field to extract location of the variant. Available options: '
                               '{%(choices)s}. Default: "HGVSc".')

    parent_parser.add_argument('-a', '--aggregate_classes', help='Aggregate specific variant classes into higher level concepts '
                               '(e.g. missense, synonymous variants will be evaluated as coding variants).',
                               action='store_true')
    
    parent_parser.add_argument('-v', '--top_vep_consequence', metavar='', default='gene_body', choices=("first", "gene_body", "smallest_offset"), 
                               help='How to select the top VEP consequence for each variant. Available options: {%(choices)s}. '
                               'Default: "gene_body": First consequence occurring in the body of a gene is selected. '
                               'If "first" is set, first consequence is selected. '
                               'If "smallest_offset" is set, consequence within body genes with the smallest offset is selected (e.g. exonic variants first, then intronic variants closest to a splice site).')
    
    parent_parser.add_argument('-i', '--do_intronic_analysis', help='Perform additional analysis of '
                               'intronic variants extracted from '
                               'HGVSc field in a bin-based faction. '
                               'Bins are attributed based on the distance '
                               'of the variant to the nearest splice junction',
                               action='store_true')

    parent_parser.add_argument('--split_splice_sites', help='When "--do_intronic_analysis" is set, ' 
                               'perform separate analysis for donor and acceptor associated variants.',
                               action='store_true')
        
    parent_parser.add_argument('-af', '--allele_frequency', metavar='', default="gnomADg_AF",
                               help='VCF field (within VEP annotations, or INFO field) '
                               'that measures frequency of the variant in a '
                               'population. If it exists in the input data, additional '
                               'plots will be drawn. If absent, VETA will simply ignore it. '
                               'Default: "gnomADg_AF". Note: Missing data will be treated as '
                               'if the variant is absent in population (converted to 0).')

    parent_parser.add_argument('--skip_heatmap', action="store_true",
                               help="Skip heatmap generation for performance analysis. "
                               "If the dataset is large, this step takes quite a while.")

    parent_parser.add_argument('--config', default=CONFIG_PATH,
                               help='Path to the config file that maps tools to the corresponding '
                               'VCF annotation. Default: \'tools_config.txt\' file '
                               'in the src/config github directory')
    # Subparsers based on parent
    benchmark_parser = subparsers.add_parser("benchmark", help='Benchmark prediction tools based on labelled data.',
                                             parents=[parent_parser])

    benchmark_parser.add_argument(dest='dataset',
                                  help='Path to the reference dataset. If a single file is '
                                       'given, veta expects the dataset to be from clinvar. If a previous '
                                       'run of VETA was performed, a cached file with the '
                                       'addition of \'tsv\' will exist, thus making processed clinvar data '
                                       'to be loaded much quicker. If a directory is '
                                       'provided instead, veta looks for labelled files within it. The '
                                       '\'*benign*\' (or \'*neutral*\') and \'*pathogenic*\' (or \'*deleterious*\') '
                                       'tags must exist in their names.')

    benchmark_parser.add_argument('-c', '--clinvar_stars', metavar='', default='1s_l',
                                  help='Level of filtering when dataset refers to the clinvar database. '
                                       'By default, a high confidence clinvar subset (3 stars with likely '
                                       'annotations) is used for performance evaluation and reference '
                                       'threshold analysis (if --do_threshold_analysis is True). '
                                       'Default: "1s_l". All the possible filtering levels are visible '
                                       'with the argument \'--listClinvarLevels\'.')

    benchmark_parser.add_argument('--omim_ids', nargs='+', help='When dataset refers to the clinvar database, '
                                  'selects variants belonging to the provided OMIM ids.')
    
    benchmark_parser.add_argument('--medgen_ids', nargs='+', help='When dataset refers to the clinvar database, '
                                  'selects variants belonging to the provided MedGen ids.')
    
    benchmark_parser.add_argument('--mondo_ids', nargs='+', help='When dataset refers to the clinvar database, '
                                  'selects variants belonging to the provided MONDO ids.')
        
    benchmark_parser.add_argument('--do_threshold_analysis', action="store_true",
                                  help="Enable reference thresholds analysis for the input dataset."
                                       "Default: False")

    benchmark_parser.add_argument('--bootstrapping', action="store_true",
                                  help="Enable bootstrapping analysis when '--do_threshold_analysis' is set."
                                       "Default: False")
    
    benchmark_parser.add_argument('--do_machine_learning', action="store_true",
                                  help="Enable machine learning analysis based on the "
                                       "tools scores to inspect the best predictors and "
                                       "create an ensemble classifier that combines multiple"
                                       " scores. Default: False.")

    predict_parser = subparsers.add_parser("interrogate", help='Evaluate predictions from unlabelled variants.',
                                           parents=[parent_parser])

    predict_parser.add_argument(
        dest='vcf', help="VCF file to evaluate tools predictions.")

    predict_parser.add_argument('--plot_these_tools', metavar='tool_name', nargs='+',
                                help="Plot scores distribution for the given tools.")

    predict_parser.add_argument('--labels', metavar='label_type',
                                choices=("Benign", "Neutral", "Pathogenic",
                                         "Deleterious", "Functional"),
                                help="If VCF represents a list of labelled variants, "
                                     "additional metrics will be inferred. This "
                                     "argument asks for the label type of input vcf. "
                                     "Available options: {%(choices)s}. Default: VCF "
                                     "does not have variants with known labels.")
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()

    clinvar_stars = ['1s', '2s', '3s', '4s', '1s_l',
                     '2s_l', '3s_l', '0s', '0s_l']
    ##############################
    ## Argparse args processing ##
    ##############################
    if args.command == "interrogate":

        assert args.metric not in ['F1', 'weighted_F1', 'mcc', 'norm_mcc', 'weighted_norm_mcc'], "Metric provided is not valid for interrogate mode."
     
        PredictionsEval(args.vcf,
                        args.out_dir,
                        args.scopes_to_evaluate,
                        args.types_of_variant,
                        args.metric,
                        args.location,
                        args.aggregate_classes,
                        args.top_vep_consequence,
                        args.do_intronic_analysis,
                        args.plot_these_tools,
                        args.labels,
                        args.allele_frequency,
                        args.skip_heatmap,
                        args.config,
                        interrogate_mode=True
                        )

    elif args.command == "benchmark":
        if args.clinvar_stars not in clinvar_stars:
            logging.info("Error. Set a valid value for the '--clinvar_stars' argument. Possible options "
                         "(value with a simple description):")
            print_clinvar_levels()

        assert args.clinvar_stars in clinvar_stars, "Set a valid value for the '--clinvar_stars' argument."
        phenotype_ids = [args.omim_ids, args.medgen_ids, args.mondo_ids]
        BenchmarkTools(args.dataset,
                       args.out_dir,
                       args.scopes_to_evaluate,
                       args.types_of_variant,
                       args.metric,
                       args.location,
                       args.aggregate_classes,
                       args.top_vep_consequence,
                       args.do_intronic_analysis,
                       args.split_splice_sites,
                       args.clinvar_stars,
                       phenotype_ids,
                       args.do_threshold_analysis,
                       args.bootstrapping,
                       args.do_machine_learning,
                       args.allele_frequency,
                       args.skip_heatmap,
                       args.config)


if __name__ == '__main__':
    main()



import logging
import os
import sys
import pandas as pd
from preprocessing.location import *

def remove_clinvar_useless(df: pd.DataFrame):
    """
    Removes conflicting, missing and
    uncertain interpretations

    :param pd.DataFrame df: Input df
    :return pd.DataFrame: Df with variants
    following the target significance values
    """
    return df.loc[df['CLNSIG'].isin(['Pathogenic', 'Benign',
                                     'Likely_pathogenic', 'Likely_benign',
                                     'Pathogenic/Likely_pathogenic', 'Benign/Likely_benign'])]


def get_clinvar_cached(fname: str):
    """
    Reads cached processed clinvar file if
    it exists. Else, processes original VCF

    :param str fname: Path to the input File
    :return pd.DataFrame: Processed clinvar df
    """
    tsv_file = fname if fname.endswith('tsv') else fname + '.tsv'

    #If clinvar was processed before
    if os.path.exists(tsv_file):
        logging.info('Cached clinvar file found. VETA will pick from that.' 
                     'If you want to run veta with different configurations, please delete the {} file and run again.'.format(tsv_file))
        return pd.read_csv(tsv_file, low_memory=False)

    return

def filter_by_condition(df: pd.DataFrame, ids:list):
    """
    Filter Clinvar database by a subset of OMIM/MedGen/MONDO IDs
    
    :param list ids: IDs from different databases.
    
    :return pd.DataFrame: Returns the variants in which at least one 
    ID in one database was found
    """

    if all(x is None for x in ids):
        return df
    
    def _get_matched_ids(df, db_field, db_ids):
        """
        Returns Clinvar IDs that match the given condition ID
        """
        re = "[0-9]+" if db_field == "MONDO:" else "[0-9A-Za-z]+"
       
        _ids = df["CLNDISDB"].str.extractall("({}{})".format(db_field, re)).iloc[:, 0].apply(lambda x: x.replace('{}'.format(db_field), '').rstrip())
        match_idx = _ids[_ids.isin(db_ids)].index.get_level_values(0)
        _df = df.loc[match_idx, :]

        return _df.id.to_list()

    ids_map = {0: 'OMIM:', 1: 'MedGen:', 2: 'MONDO:'}
    out_ids = []

    for i, db_ids in enumerate(ids):
        if db_ids is None:
            continue
        
        db_field = ids_map[i]
        out_ids.extend(_get_matched_ids(df, db_field, db_ids))
 
    df = df[df.id.isin(out_ids)]
    if df.shape[0] < 1:
        raise ValueError('No variants left after filtering by {} condition ID(s).'.format(ids))
    else:
        logging.info('Number of variants after filtering by condition ID(s) ({}): {}'.format([x for x in ids if x is not None], df.shape[0]))
   
    return df 

def filter_clinvar_sure(df: pd.DataFrame):
    """
    Removes variants with likely assignments
    """
    return df[df['CLNSIG'].isin(['Pathogenic', 'Benign'])].copy(deep=True)


def filter_clinvar_1_star(df: pd.DataFrame):
    """
    Removes variants with less than 1 star
    """
    # One submitter provided an interpretation with
    # assertion criteria (criteria provided, single submitter)
    # or multiple submitters provided assertion criteria but there
    # are conflicting interpretations. In such cases
    # the independent values are enumerated for clinical
    # significance (criteria provided, conflicting interpretations)
    to_remove = ['no_assertion_criteria_provided',
                 'no_assertion_provided',
                 'no_interpretation_for_the_single_variant']

    return df[~df['CLNREVSTAT'].isin(to_remove)].copy()


def filter_clinvar_2_stars(df: pd.DataFrame):
    """
    Removes variants with less than 2 stars
    """
    return df[(df["CLNREVSTAT"].str.contains('criteria_provided,_multiple_submitters,_no_conflicts')) |
              (df["CLNREVSTAT"].str.contains('reviewed_by_expert_panel')) |
              (df["CLNREVSTAT"].str.contains('practice_guideline'))].copy()


def filter_clinvar_3_stars(df: pd.DataFrame):
    """
    Removes variants with less than 3 stars
    """

    _df = df[df.CLNREVSTAT.str.contains('reviewed_by_expert_panel')].copy()
    return _df


def filter_clinvar_4_stars(df: pd.DataFrame):
    """
    Removes variants with less than 4 stars
    """
    return df[(df["CLNREVSTAT"].str.contains('practice_guideline'))].copy()




import os
import pandas as pd
import matplotlib.patches as mpatches
import numpy as np
from .plots_utils import *
plt.switch_backend('agg')


def plot_bootstrap_thresholds(data: pd.Series,
                               tool: str, 
                               beta: str,
                               location: str,
                               new_thresh: float,
                               reference_thresh: float,
                               ci_down_95: float,
                               ci_up_95:float,
                               outdir: str):
    """
    :param str data: Best thresholds at each bootstrap sample
    :param str tool: Name of the tool
    :param str beta: Beta value
    :param str location: Location filter for the variants analysed
    :param np.array threshold_range: Observed range of scores for the tool
    :param float new_thresh: Best new thresh after bootstratping
    :param float reference_threshold: Reference threshold for the tool
    :param float ci_down_95: Lower bound 95% CI
    :param float ci_up_95: Upper bound 95% CI
    :param str outdir: Output directory
    """
    outdir = os.path.join(outdir, "plots/bootstrapping")
    os.makedirs(outdir, exist_ok=True)
    
    plt.hist(data, bins=15, density=True)
    plt.axvline(x=reference_thresh, color='brown', linestyle="--", linewidth=2, label="Reference threshold")
    plt.axvline(x=new_thresh, color='darkgreen', linestyle="--", linewidth=2, label="New threshold")
    plt.axvline(x=ci_down_95, color='k', linestyle=':', linewidth=2, label="95% Bootstrap percentile")
    plt.axvline(x=ci_up_95, color='k', linestyle=':', linewidth=2)
    plt.title('Distribution of best {} thresh for beta={}'.format(tool, beta))
    plt.xlabel('Threshold range')
    plt.ylabel('Probability')
    plt.legend(bbox_to_anchor=(1.1, 1.05))

    f_name = os.path.join(outdir, "threshold_{}_{}_f{}.pdf".format(tool.replace(" ", "_"),
                                                                 location,
                                                                 beta))
    plt.savefig(f_name, bbox_inches='tight')
    plt.close()

def plot_optimal_thresholds(tool: str, location: str, threshold_range: np.array,
                            accuracies: list, sensitivities: list, specificities: list,
                            beta_values: dict,
                            reference_threshold: float,
                            new_thresholds: dict,
                            outdir: str,
                            simple=False):
    """
    :param str tool: Tool to plot new thresholds
    :param str location: Location filter for the variants analysed
    :param np.array threshold_range: Observed range of scores for the tool
    :param list accuracies: Accuracy at each threshold value
    :param list sensitivities: Sensitivity/recall at each threshold value
    :param list specificities: Specificity at each threshold value
    :param dict beta_values: Fbeta scores at each threshold value for each
        beta values used in the analysis
    :param float reference_threshold: Reference threshold for the tool
    :param dict new_thresholds: Optimized threshold at each beta value
    :param str outdir: Output directory
    :param boolean simple: Draw simpler plot. Default: `False`. All new
        recommended thresholds at each beta value will be plotted.
    """

    plt.figure(figsize=(8, 6))
    plt.plot(threshold_range, accuracies, label="Accuracy", linewidth=2, color='#B22400')
    plt.plot(threshold_range, sensitivities, label="Sensitivity", linewidth=2, color='darkmagenta')
    plt.plot(threshold_range, specificities, label="Specificity", linewidth=2, color='#429C40')
    plt.axvline(x=reference_threshold, color='k', linestyle=':', linewidth=3, label="Recommended threshold")

    patches = []
    if not simple:
        col = sns.dark_palette("#69d", len(beta_values))
        col_i = 0
        for k in sorted(beta_values.keys()):
            plt.plot(threshold_range, beta_values[k], color=col[col_i], linestyle="--", linewidth=1)
            plt.axvline(x=new_thresholds[k], color=col[col_i], linestyle="--", linewidth=1)
            patches.append(mpatches.Patch(color=col[col_i], label='+1 TP = +{} FP'.format(k), linestyle="--"))
            col_i += 1

    plt.subplots_adjust(right=0.7) 
    leg1 = plt.legend(bbox_to_anchor=(1, 0.7), frameon=True)

    if not simple:
        plt.legend(handles=patches, title="New thresholds", bbox_to_anchor=(1, 1))
    plt.gca().add_artist(leg1)
    plt.title("{}".format(tool))
    plt.xlabel("Threshold values")
    plt.ylabel("Metrics")

    outdir = os.path.join(outdir, 'plots')
    os.makedirs(outdir, exist_ok=True)
    f_name = os.path.join(outdir, "threshold_{}_{}{}.pdf".format(tool.replace(" ", "_"),
                                                                 location,
                                                                 simple and "_simple" or ""))
    plt.savefig(f_name, bbox_inches='tight')
    plt.gcf().clear()
    plt.clf()
    plt.cla()
    plt.close()



import os
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import logging
plt.switch_backend('agg')
cmap = sns.diverging_palette(220, 10, as_cmap=True)
from matplotlib.lines import Line2D
import matplotlib.colors as c
import matplotlib.patches as mpatches
from statannotations.Annotator import Annotator
from plots.plots_utils import *


def plot_density_by_class(data: pd.DataFrame,
                          thresholds: list,
                          fname: str,
                          min_predicted: int = 20):
    """
    Plot class distribution for a given tool taking
    into account the reference threshold
    :param pd.DataFrame data: Scores and class
        for a single tool
    :param list thresholds: Reference thresholds
    :param str fname: Output file name
    :param int min_predicted: Minimum number of variants
    predicted to draw plots. Default: `20`
    """
    os.makedirs(os.path.dirname(fname), exist_ok=True)
    
    fraction_nan = round(np.sum(np.sum(data.iloc[:, 0].isnull())) / data.shape[0] * 100, 3)
    df_ = data[~data.iloc[:, 0].isnull()]

    if df_.shape[0] > min_predicted:

        sns.kdeplot(x=df_.columns[0],
                    data=df_,
                    palette={True: 'firebrick', False: 'royalblue'},
                    shade=True,
                    linewidth=1.5,
                    hue='label',
                    legend=False,
                    warn_singular=False)

        plt.axvline(x=[x[2] for x in thresholds if x[0] == list(df_)[0]][0], color='k', linestyle="--")
        legend_d = {'Pathogenic': 'firebrick',
                    'Benign': 'royalblue',
                    'Reference threshold': 'black'}

        handles = []
        for _l, _c in legend_d.items():
            style = '--' if _l not in ['Benign', 'Pathogenic'] else '-'

            handles.append(Line2D([], [], color=_c, linestyle=style,
                                  markersize=10,
                                  markeredgewidth=1.5,
                                  label=_l))
        plt.legend(handles=handles)

        plt.xlabel(list(df_)[0] + " ({}% missing data)".format(fraction_nan))
        ends = (max(df_.iloc[:, 0]) - min(df_.iloc[:, 0])) * 0.05
        plt.xlim(min(df_.iloc[:, 0]) - ends, max(df_.iloc[:, 0]) + ends)
        plt.savefig(fname + ".pdf")
        plt.close()


def plot_allele_frequency(df: pd.DataFrame,
                          fname: str,
                          af_col: str = "gnomADg_AF"):
    """
    Plots allele frequencies for each class

    :param pd.DataFrame df: Input df
    :param str fname: Output basemame
    :param str af_col: Column name that accounts
        for allele frequencies. Default: `gnomAD_genomes`.
        If column does not exist, analysis will be skipped.
    """

    if af_col not in df.columns:
        return
    else:
        os.makedirs(os.path.dirname(fname), exist_ok=True)
        
    df['grouper'] = df['outcome'].astype(str) + '\nN = ' + df['count_class'].astype(str)
    order = sorted(list(df['grouper'].unique()))

    ax = sns.boxplot(data=df, x="grouper", order=order, y=af_col)
    try:
        if len(order) == 2:
            annotator = Annotator(ax, 
                                pairs=[(order[0], order[1])], 
                                data=df,
                                x='grouper', 
                                y=af_col, 
                                order=order)
            annotator.configure(test='Mann-Whitney',
                                text_format='star',
                                loc='inside',
                                verbose=0,
                                pvalue_format_string='{:.4f}')
            annotator.apply_and_annotate()

        plt.xlabel("")
        plt.ylabel("Allele frequency")
        plt.tight_layout()
        plt.savefig(fname)
        plt.close()
    except ValueError:
        plt.close()
        pass


def plot_unscored(data: pd.DataFrame, fname: str):
    """
    Plot fraction of unscored variants by each
    tool in `data`.

    :param pd.DataFrame data: Df with stats
        for each tool
    :param str fname: Output file
    """
    _data = data[['tool', 'mp', 'mn', 'fraction_nan', 'total']].copy()

    _data.loc[:, 'Pathogenic'] = round(data.mp / data.total, 3)
    _data.loc[:, 'Benign'] = round(data.mn / data.total, 3)
    _data = _data[['tool', 'Pathogenic', 'Benign', 'fraction_nan']].set_index('tool')

    os.makedirs(os.path.dirname(fname), exist_ok=True)

    plt.figure()
    plt.rcParams.update({'font.size': 10}) 
    ax = _data.sort_values('fraction_nan').drop('fraction_nan', axis=1).plot.barh(stacked=True, 
                                                                                  width=1,
                                                                                  color={'Pathogenic': 'darkred', 'Benign': 'skyblue'},
                                                                                  edgecolor='k',
                                                                                  alpha=0.7,
                                                                                  fontsize=8,
                                                                                  figsize=(4.25, 4.5))
    ax.legend(fontsize=6)
    ax.set(xlabel='Fraction of unscored variants', ylabel='')
    plt.xlim(0, 1)
    plt.tight_layout()
    plt.savefig(fname)
    plt.close()


def plot_metrics(data: pd.DataFrame, fname: str, metric: str):
    """
    Plot tools ranked by a given metric

    :param pd.DataFrame data: Df with stats
        for each tool
    :param str fname: Output name
    :param str metric: Metric used
        to rank the tools
    """
    if data.empty:
        return 
    os.makedirs(os.path.dirname(fname), exist_ok=True)

    try:
        data = data[data.fraction_nan < 0.95]
    except AttributeError:
        pass

    data = data.sort_values(metric)

    my_range_coverage = list(range(1, len(data.index) + 1))
    my_range_specificity = np.arange(1 - 0.2, len(data.index)).tolist()
    my_range_sensitivity = np.arange(1 + 0.2, len(data.index) + 0.5).tolist()
    
    n_tools = data.shape[0]
    if n_tools <= 10:
        figsize = (6 , 3.4)
        _bbox = 1.6
        _left = 0.25
        _right = 0.75
        m_title_l = 0.03
        
    elif n_tools < 20:
        figsize = (6.3 , 4.4)
        _bbox = 1.4
        _left = 0.2 
        _right = 0.8
        m_title_l = 0.03
    elif 20 <= n_tools <= 30:
        figsize = (7.5 , 6.5)
        _bbox = 1.3
        _left = 0.3
        _right = 0.8
        m_title_l = 0.04
    else:
        figsize = (8.2, 7)
        _bbox = 1.35
        _left = 0.3
        _right = 0.8
        m_title_l = 0.05
    fig, ax = plt.subplots(figsize=figsize)
    
    if n_tools > 10:
        plt.text(m_title_l, 0.9, "Tool({})".format(metric), fontsize=10, transform=plt.gcf().transFigure)

    _target_col = "precision" if metric in ['F1', 'weighted_F1'] else "specificity"

    plt.scatter(data[_target_col], my_range_specificity,
                color='skyblue',
                alpha=1,
                marker='s',
                edgecolors='black',
                linewidths=0.5,
                label=_target_col.capitalize())

    plt.scatter(data['sensitivity'], my_range_sensitivity,
                color='yellow',
                alpha=0.75,
                marker='v',
                edgecolors='black',
                linewidths=0.5,
                label='Sensitivity')

    plt.scatter(data['coverage'],
                my_range_coverage,
                color='brown',
                alpha=0.75,
                marker='o',
                edgecolors='black',
                linewidths=0.5,
                label='Fraction_predictable')
    
    rectangule_attr = [_target_col, 'sensitivity', 'coverage']
    i = 0
    for idx, rows in data.iterrows():
        val_range = list(rows[rectangule_attr])
        width = [min(val_range), max(val_range)]
        height = 1
        ax.add_patch(
            plt.Rectangle(xy=(width[0] - 0.01, my_range_coverage[i] - 0.4),
                          width=width[1] - width[0] + 0.02,
                          height=height,
                          linewidth=0.75,
                          color='gray',
                          fill=False))
        i += 1
    
    ax.grid(axis='x', linestyle='dashed')
    plt.subplots_adjust(left=_left, right=_right)
    plt.legend(loc="upper right",
               bbox_to_anchor=(_bbox, 1),
               borderaxespad=0,
               prop=dict(size=8))

    if n_tools <= 10:
        plt.yticks(my_range_coverage, data['tool'] + " (" + data[metric].astype(str) + ")", fontsize=10)
    else:
        plt.yticks(my_range_coverage, data['tool'] + " (" + data[metric].astype(str) + ")")

    if all(col in data.columns for col in ['total', 'total_p', 'total_n']):
        plt.title("#variants: {} ({} pos, {} neg)".format(data["total"].iloc[0],
                                                          data["total_p"].iloc[0],
                                                          data["total_n"].iloc[0]))

    plt.tight_layout()
    plt.savefig(fname, bbox_inches='tight')
    plt.close()


def plot_tools_total_correct(data: pd.DataFrame, fname: str):
    """
    Plots the total number of correct
    classification and ranks the display
    by the top tool

    :param pd.DataFrame data: Df with stats
        for each tool
    :param str fname: Output file
    """
    os.makedirs(os.path.dirname(fname), exist_ok=True)
    
    df = data.iloc[::-1]
    set_style()
    ind = np.arange(df.shape[0])
    df.sort_values(by=['correct'], inplace=True)
    w = 0.8

    fig, ax = plt.subplots()
    plt.xlim(df['total'][0])
    ax.barh(ind, df['correct'],
            align='center',
            color='green',
            zorder=10,
            alpha=0.8,
            height=w,
            edgecolor="black",
            linewidth=0.75)

    ax.set(yticks=ind, yticklabels=df['tool'])
    ax.grid(True)
    ax.margins(0.00)
    ax.invert_xaxis()
    ax.tick_params(axis='both', which='major', labelsize=8)
    ax.set_xlabel('# Number of correctly predicted variants')
    for tick in ax.xaxis.get_minor_ticks():
        tick.tick1line.set_markersize(0)
        tick.tick2line.set_markersize(0)

    set_size(fig, len(df['tool']))
    fig.tight_layout()
    plt.savefig(fname)
    plt.close()


def plot_tools_barplot_only_correct(data: pd.DataFrame, fname: str):
    """
    Plot tools barplot for correct
    prediction alone separated by
    class labels where tools are
    sorted by the correct number of
    pathogenic variants identified

    :param pd.DataFrame data: Df with stats
        for each tool
    :param str fname: Output name
    """
    os.makedirs(os.path.dirname(fname), exist_ok=True)
    df = data.iloc[::-1]
    df = df[df.fraction_nan < 0.95]
    set_style()

    w = 0.5
    fig, axes = plt.subplots(ncols=2, sharey='all')
    df.sort_values('tp', inplace=True, ascending=True)
    ind = np.arange(df.shape[0])

    axes[0].barh(ind, df['tp'],
                 align='center',
                 color='green',
                 zorder=10,
                 alpha=0.8,
                 height=w,
                 edgecolor='black',
                 linewidth=0.75)

    axes[1].barh(ind, df['tn'],
                 align='center',
                 color='green',
                 zorder=10,
                 alpha=0.8,
                 height=w,
                 edgecolor='black',
                 linewidth=0.75)

    try:
        axes[0].set_xlim(0, df["tp"][0] + df["fn"][0])
        axes[1].set_xlim(0, df["tn"][0] + df["fp"][0])
    except KeyError:
        pass

    axes[0].set(yticks=ind, yticklabels=df['tool'])
    axes[0].set_xlabel('# Pathogenic Variants')
    axes[1].set_xlabel('# Benign Variants')

    for ax in axes.flat:
        ax.grid(False)
        ax.margins(0.00)
        ax.tick_params(axis='both', which='major', labelsize=8)
        for tick in ax.xaxis.get_minor_ticks():
            tick.tick1line.set_markersize(0)
            tick.tick2line.set_markersize(0)

    set_size(fig, len(df['tool']))
    fig.tight_layout()
    plt.savefig(fname)
    plt.close()


def plot_tools_barplot(data: pd.DataFrame, fname: str, metric: str):
    """
    Tools performance barplot with two
    axis (pathogenic and benign).

    :param pd.DataFrame data: Input df
    :param str fname: Output basename
    :param str metric: Metric to evaluate
    """
    if data.empty:
        return
    
    os.makedirs(os.path.dirname(fname), exist_ok=True)
    
    # revert order
    df = data.iloc[::-1]
    df = df[df.fraction_nan < 0.95]
    set_style()
    w = 0.8

    fig, axes = plt.subplots(ncols=2,
                             sharey='all',
                             figsize=(15, 15))
    ind = np.arange(df.shape[0])

    # Pathogenic variants
    axes[0].barh(ind, df['tp'],
                 align='center',
                 color='darkblue',
                 alpha=0.7,
                 zorder=10,
                 height=w,
                 edgecolor='black',
                 linewidth=0.75)

    axes[0].barh(ind, df['fn'],
                 left=df['tp'],
                 align='center',
                 alpha=0.8,
                 color='darkred',
                 zorder=10,
                 height=w,
                 edgecolor='black',
                 linewidth=0.75)

    axes[0].barh(ind, df['mp'],
                 left=df['scored_p'],
                 align='center',
                 color='lightgrey',
                 zorder=10,
                 height=w,
                 edgecolor='black',
                 alpha=0.9,
                 linewidth=0.75)

    # Benign variamts
    axes[1].barh(ind, df['tn'],
                 align='center',
                 color='darkblue',
                 alpha=0.7,
                 zorder=10,
                 height=w,
                 edgecolor='black',
                 linewidth=0.75)

    axes[1].barh(ind, df['fp'],
                 left=df['tn'],
                 align='center',
                 alpha=0.8,
                 color='darkred',
                 zorder=10,
                 height=w,
                 edgecolor='black',
                 linewidth=0.75)

    axes[1].barh(ind, df['mn'],
                 left=df['scored_n'],
                 align='center',
                 color='lightgrey',
                 zorder=10,
                 height=w,
                 edgecolor='black',
                 alpha=0.9,
                 linewidth=0.75)

    axes[0].invert_xaxis()
    axes[0].set(yticks=ind,
                yticklabels=["{} ({:.2f})".format(t, p) for
                             t, p in df[['tool', metric]].values.tolist()])

    axes[0].set_xlabel('# Pathogenic Variants')
    axes[1].set_xlabel('# Benign Variants')
    for ax in axes.flat:
        ax.grid(False)
        ax.margins(0.00)
        ax.tick_params(axis='both', which='major', labelsize=8)
        for tick in ax.xaxis.get_minor_ticks():
            tick.tick1line.set_markersize(0)
            tick.tick2line.set_markersize(0)

    colors = ["darkblue", "darkred", "lightgrey"]
    patches = [mpatches.Patch(facecolor=c, edgecolor=c) for c in colors]
    axes[1].legend(patches, ["Correct", "Incorrect", "Unpredictable"],
                   bbox_to_anchor=(1.7, 1),
                   borderaxespad=0,
                   loc='upper right',
                   prop=dict(size=8))

    set_size(fig, len(df['tool']))
    fig.tight_layout()
    plt.savefig(fname)
    plt.close()
    sns.reset_defaults()


def plot_curve(data: list,
               fname: str,
               class_counts: tuple,
               is_roc: bool = True,
               max_nan_allowed: float = 0.5):
    """
    Plot ROC or pr curves for all tools in data

    :param list data: ROC analysis results for
    all the tools
    :param str fname: Output basename
    :param tuple class_counts: Number of positive and
    negative variants
    :param bool is_roc: Whether analysis refers to
    ROC curve. If `False`, precision-recall curves are
    drawn. Default: `True`
    :param float max_nan_allowed: Maximum
    fraction of missing variants per tool allowed
    for the curve to be drawn. Default: `0.5`
    """
    os.makedirs(os.path.dirname(fname), exist_ok=True)
 
    data = [x for x in data if len(x) > 0]
    
    if data:
        if is_roc:
            colnames = ['tool', 'fraction_nan', 'label', 'thresholds', 'True Positive Rate (TPR)',
                        'False Positive Rate (FPR)', 'auROC']
            to_explode = ['thresholds', 'True Positive Rate (TPR)', 'False Positive Rate (FPR)']
        else:
            colnames = ['tool', 'fraction_nan', 'label', 'thresholds', 'Recall', 'Precision', 'auPRC']
            to_explode = ['thresholds', 'Recall', 'Precision']

        df_metrics = pd.DataFrame.from_records(data, columns=colnames)
        df_metrics = df_metrics[df_metrics.thresholds.notna()]
        df_metrics = df_metrics.explode(to_explode).reset_index()

        if is_roc:
            df_metrics['True Positive Rate (TPR)'] = pd.to_numeric(df_metrics['True Positive Rate (TPR)'])
            df_metrics['False Positive Rate (FPR)'] = pd.to_numeric(df_metrics['False Positive Rate (FPR)'])
            df_metrics["tool_with_auROC"] = df_metrics["label"] + " auROC=" + \
                                            df_metrics["auROC"].round(3).map(str) + ")"
            hue = "tool_with_auROC"
            x = "False Positive Rate (FPR)"
            y = "True Positive Rate (TPR)"
            df_metrics = df_metrics.sort_values('auROC', ascending=False)
            
            no_min_scored = df_metrics[df_metrics['fraction_nan'] > max_nan_allowed].tool.unique().tolist()
            if no_min_scored:
                logging.info('ROC curves will not be drawn for the following tool(s) (more than {} fraction of missing predictions): {}'.format(max_nan_allowed, ','.join(no_min_scored)))
        else:
            df_metrics['Recall'] = pd.to_numeric(df_metrics['Recall'])
            df_metrics['Precision'] = pd.to_numeric(df_metrics['Precision'])
            df_metrics["tool_with_auPRC"] = df_metrics["label"] + " auPRC=" + \
                                            df_metrics["auPRC"].round(3).map(str) + ")"
            hue = "tool_with_auPRC"
            x = "Recall"
            y = "Precision"
            df_metrics = df_metrics.sort_values('auPRC', ascending=False)

        df_metrics = df_metrics[df_metrics['fraction_nan'] <= max_nan_allowed]

        # Since S-CAP has several different reference
        # threshold, S-CAP is removed from these analyses
        df_metrics = df_metrics[~df_metrics.tool.str.contains("S-CAP")]

        # If many tools to plot, change color pallette
        if df_metrics.tool.unique().size > 10:
            sns.set(rc={'figure.figsize':(6, 4)})
            sns.set_palette(sns.mpl_palette("magma", df_metrics.tool.unique().size))
            fontsize="small"
        else:
            sns.set(rc={'figure.figsize':(6, 4)})
            sns.set_palette(sns.color_palette("colorblind"))
            fontsize="medium"
        
        sns.set_style("white")
        ax = sns.lineplot(x=x, y=y,
                        data=df_metrics,
                        hue=hue)
        ax.set_aspect(1)
        sns.despine()
        plt.title("N pos = {}; N neg = {}".format(class_counts[0], class_counts[1]))
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., frameon=False, fontsize=fontsize)
        plt.ylim(0, 1.05)
        plt.tight_layout()

        plt.savefig(fname, bbox_inches='tight')
        plt.close()
        sns.reset_defaults()
        
        _metric = 'auROC' if is_roc else 'auPRC'   
        df_metrics = df_metrics[['tool', _metric]].drop_duplicates()
        return dict(zip(df_metrics.tool, df_metrics[_metric])) 

###############################
## Heatmap related functions ##
###############################
GT_LABEL = '(*)True labels'


def prepare_dataset_for_heatmap(df):
    """
    Convert boolean dataframe into numbers
    It discards info about each variant

    :param pd.DataFrame df: Boolean dataframe to process
    :return pd.DataFrame: Dataframe ready for plotting
    """
    df['blank'] = pd.Series(np.nan, index=np.arange(df.shape[0]))
    df = df[['label'] + [col for col in df.columns if '_prediction' in col]].copy()
    df.columns = [GT_LABEL] + [col.replace("_prediction", "") for col in df.columns if '_prediction' in col]
    f = lambda x: pd.isna(x) and np.nan or int(x)
    for col in df.columns:
        df[col] = df[col].apply(f)
        if df[col].isnull().all():
            del df[col]
    df = df.fillna(-1)
    df = df.reset_index(drop=True)
    df = df.sort_values(by=GT_LABEL)
    return df


def plot_heatmap(df,
                 fname: str,
                 cluster_rows: bool = False,
                 skip_preparation: bool = False):
    """
    Plot a binary heatmap of tools performance

    :param pd.DataFrame df: Boolean dataframe with predictions after
        running `apply_thresholds` method
    :param str output_dir: Output directory
    :param bool cluster_rows: Cluster by rows. Default: `False`
    :param bool skip_preparation: Whether input dataset is ready. Default: `False`
    """
    os.makedirs(os.path.dirname(fname), exist_ok=True)
    
    if len(df.columns) < 2:
        return

    if skip_preparation is False:
        df = prepare_dataset_for_heatmap(df)

    colors = {"ivory": -1, "lightsteelblue": 0, "darksalmon": 1}
    l_colors = sorted(colors, key=colors.get)
    cmap = c.ListedColormap(l_colors)

    cm = sns.clustermap(df, row_cluster=cluster_rows, cmap=cmap,
                        linecolor='black',
                        linewidths=0.0,
                        cbar=False,
                        annot=False,
                        xticklabels=1,
                        yticklabels=False)

    # Aesthetics of the clustergrid
    cm.ax_row_dendrogram.set_visible(False)
    cm.ax_col_dendrogram.set_visible(True)
    cm.cax.set_visible(False)
    ax = cm.ax_heatmap
    ax.set_xticklabels(ax.get_xticklabels(), rotation=80)

    # Add legend
    legend_patch = [mpatches.Patch(facecolor=c, edgecolor='black', linewidth=0.5) for c in
                    ['darksalmon', 'lightsteelblue', 'ivory']]

    legend = ax.legend(legend_patch,
                       ['Pathogenic', 'Benign', 'Unpredictable'],
                       loc='upper right',
                       bbox_to_anchor=(1.25, 1),
                       frameon=True,
                       handlelength=1,
                       handleheight=1)
    legend.get_frame().set_edgecolor('black')
    legend.get_frame().set_linewidth(1.0)

    # Add vertical lines
    ax.vlines(range(0, df.shape[1]), ymin=0, ymax=df.shape[0], colors='black', linewidths=0.5)
    ax.axhline(y=0, color='k', linewidth=5)
    ax.axhline(y=df.shape[0], color='k', linewidth=5)
    ax.axvline(x=0, color='k', linewidth=5)
    ax.axvline(x=df.shape[1], color='k', linewidth=5)

    plt.savefig(fname, bbox_inches='tight', pad_inches=0)
    plt.close()



import logging
import multiprocessing
import os
import subprocess
import tempfile
from collections import OrderedDict, defaultdict
from functools import partial
import re
import pandas as pd
from cyvcf2 import VCF
from tqdm import tqdm
from preprocessing import osutils


def process_vcf(vcf: str,
                tools_config: dict,
                thresholds: list,
                select_conseq: str,
                af_col: str,
                is_clinvar: bool = False):
    """
    Scans a VEP annotated VCF file and extracts all the
    tools scores

    :param str vcf: VCF file
    
    :param dict tools_config: Dict with the
        map between available tools and VCF
        annotations (already parsed, only
        existing fields should be passed here).
        
    :param list thresholds: List of tools
        to process based on the tools config file
        as well as the selected scope for the
        analysis
        
    :param str select_conseq: Strategy to select
    a single consequence block per variant 
    
    :param str af_col: VCF field that measures
            population frequency of variants.
    :param bool is_clinvar: Whether `vcf` is
        from clinvar database.

    :return dict: Dict with scores keyed by variant
    :return list: List with the tools provided in the
    config that were not found in the VCF. (Will be
    used to update the config_tools variable)

    """
    logging.info("Extracting predictions from VCF.")

    # List of required INFO fields when
    # dataset is from Clinvar
    CLINVAR_FIELDS = ['CLNREVSTAT', 'CLNSIG']
    clinvar_confirmed = False

    # List of tools belonging to the
    # provided scope that will be analysed
    # Requires to be present in the updated config
    # (some tools might be removed from config after processing first file)
    TOOLS_TO_ANALYSE = [t[0] for t in thresholds if t[0] in tools_config.keys()]

    # MAP with the list of VCF fields per tool
    TOOLS_CONFIG_MAP = {k: v[0] for k, v in tools_config.items() if k in TOOLS_TO_ANALYSE}

    # List with the VCF fields to check
    # Present fields in the header will be
    # popped out sequentially
    MISSING_VCF_FIELDS_FLAT = [i for sublist in [v for _k, v in TOOLS_CONFIG_MAP.items()]
                               for i in sublist]

    # Missing tools from VCF to return
    # and exclude from analysis
    MISSING_VCF_TOOLS = []

    # Indexes is an ordered dict with the index(s)
    # where information about a given tool is stored
    # in the VEP annotation field
    vep_indexes = OrderedDict()

    # List of VCF fields (not tool names) that are not
    # found in the VEP annotation field
    absent_from_VEP_annot = []

    # Dict with the list of tools and corresponding
    # VCF fields found in the INFO field as a single
    # annotation
    present_in_INFO = defaultdict(list)

    # List of tool names with all annotation fields
    # found in the VEP annotation field
    all_vep_annotations = []
    present_in_VEP_annot = []

    # Parse header and count number of variants
    vcf_data = VCF(vcf)
    if vcf_data.contains("ANN"):
        VEP_TAG = "ANN"
    elif vcf_data.contains("CSQ"):
        VEP_TAG = "CSQ"
    else:
        raise ValueError("VETA requires VEP annotations. ANN or CSQ field was not found in "
                         "the INFO field of the input VCF file. Exiting.\n")

    for field in vcf_data.header_iter():
      
        # If field is in VEP annotations
        if field["HeaderType"] == "INFO" and field["ID"] == VEP_TAG:
            all_vep_annotations = field["Description"].split("Format:")[1][:-1].strip().split("|")
            if af_col in all_vep_annotations:
                vep_indexes[af_col] = [all_vep_annotations.index(af_col)]

            # Looks only for scores belonging
            # to the specific scope set
            # Also look for the AF column set
            for _tool in TOOLS_TO_ANALYSE:

                _tool_field = TOOLS_CONFIG_MAP[_tool]
                _present, _absent = _check_if_field_exists(_tool_field, all_vep_annotations)

                if _present:
                    _fields_present = [v[0] for v in _present]
                    present_in_VEP_annot.extend(_fields_present)
                    vep_indexes[_tool] = [v[1] for v in _present]
                    [MISSING_VCF_FIELDS_FLAT.remove(i) for i in _fields_present]

                if _absent:
                    absent_from_VEP_annot.extend(_absent)

        elif field["HeaderType"] == "INFO":
            if field['ID'] in MISSING_VCF_FIELDS_FLAT:
                _tool_name = [_t for _t, v in TOOLS_CONFIG_MAP.items() if field['ID'] in v][0]
                present_in_INFO[_tool_name].append(field['ID'])
                MISSING_VCF_FIELDS_FLAT.remove(field['ID'])

            if field['ID'] in CLINVAR_FIELDS:
                clinvar_confirmed = True

            if field['ID'] == af_col and af_col not in vep_indexes.keys() and af_col not in present_in_INFO.keys():
                present_in_INFO[af_col].append(field['ID'])

    if is_clinvar and not clinvar_confirmed:
        raise ValueError('Input VCF is not from Clinvar ({}) fields were '
                         'not found in the VCF INFO annotations. They are '
                         'required for clinical significance extraction.\n'
                         'Additionally, be aware that when running in '
                         'benchmark mode, non-Clinvar files must be passed '
                         'via an input directory where target files '
                         'referring to each class are located.'.format(CLINVAR_FIELDS))

    # if there are VCF fields from
    # the config absent in VCF header
    # LOG absent fields
    if MISSING_VCF_FIELDS_FLAT:

        for tool, fields in TOOLS_CONFIG_MAP.items():

            for _f in fields:
                if _f in MISSING_VCF_FIELDS_FLAT:
                    logging.info("\'{}\' field does not exist "
                                 "in VCF. Corresponding tool "
                                 "(\'{}\') will be discarded from "
                                 "analysis.".format(_f, tool))

                    MISSING_VCF_TOOLS.append(tool)
                    # Delete tool from analysis
                    # if at least one subfield
                    # is not present
                    try:
                        del vep_indexes[tool]
                    except KeyError:
                        pass

                    try:
                        del present_in_INFO[tool]
                    except KeyError:
                        pass

    n_variants = 0
    for _ in vcf_data:
        n_variants += 1
    vcf_data.close()

    # Pack kwargs
    args = {'VEP_TAG': VEP_TAG,
            'all_vep_annotations': all_vep_annotations,
            'present_in_INFO': present_in_INFO,
            'vep_indexes': vep_indexes,
            'is_clinvar': is_clinvar,
            'select_conseq': select_conseq}
    
    # Variant processing
    if n_variants > 5000:
        header = subprocess.Popen(["bcftools", "view", "-h", vcf],
                                  stdout=subprocess.PIPE).stdout.readlines()

        body_variants = tempfile.NamedTemporaryFile()
        subprocess.call(["bcftools", "view", "-H", vcf], stdout=body_variants)
 
        list_files = osutils.split_file_in_chunks(body_variants.name, header, n_variants)

        with multiprocessing.Pool() as p:
            df_list = list(tqdm(p.imap(partial(_iter_variants, **args), list_files)))

        logging.info("Merging data from parallel VCF processing.")
        scores = pd.concat(df_list)
        [os.remove(f) for f in list_files]
        logging.info("Done.")
        p.close()

    else:
        scores = _iter_variants(vcf, **args)
    return scores, MISSING_VCF_TOOLS


def _iter_variants(vcf: str, **kwargs):
    """
    Iterates over a VCF file

    :param str vcf: vcf file
    :param dict **kwargs: Additional info obtained
        when processing the VCF header

    :return dict: Dictionary with scores for each variant
    """
    scores = []
    keys = []

    vcf_data = VCF(vcf)

    # Iterate over VCF
    for record in vcf_data:
        
        colnames = ['index', 'chr', 'pos', 'ref', 'alt', 'id', 'type', 'subtype']
        _single_out = []
        
        # Only keep variants with VEP annotations
        vep_annotation = record.INFO.get(kwargs['VEP_TAG'])
        if not vep_annotation:
            continue
        
        if record.ID != "." and record.ID is not None:
            key = record.CHROM + "_" + str(record.POS) + "_" + str(record.ID) + "_" + str(record.ALT[0])

        else:
            key = record.CHROM + "_" + str(record.POS) + "_" + str(record.REF) + "_" + str(record.ALT[0])
        
        if key in keys:
            logging.info("Variant {},{},{},{}, {} is repeated in the VCF. "
                         "Keeping only first occurrence".format(record.CHROM,
                                                                record.POS,
                                                                record.REF,
                                                                record.ALT[0],
                                                                record.ID))
            continue

        else:
            keys.append(key)

        _single_out.append(key)

        # MNPs
        if record.var_type == "indel" and len(record.REF) == len(record.ALT[0]):
            _single_out.extend([record.CHROM,
                                record.POS,
                                record.REF,
                                record.ALT[0],
                                record.ID,
                                record.var_type, 
                                "mnp"])

        else:
            _single_out.extend([record.CHROM,
                                record.POS,
                                record.REF,
                                record.ALT[0],
                                record.ID,
                                record.var_type,
                                record.var_subtype])

        if record.POS == 44874309:
            a=1
        vep_info = _select_consequence(vep_annotation, **kwargs)
        # Add some VEP fields, if they exist
        for _field in ['Existing_variation', 'HGVSc', 'HGVSg', 'SYMBOL', 'Consequence']:
            colnames.append(_field)
            try:
                _single_out.append(vep_info[kwargs['all_vep_annotations'].index(_field)])
            except ValueError:
                _single_out.append(None)

        # Look for externally annotated scores
        # that are single fields in the INFO field
        if len(kwargs['present_in_INFO']) > 0:

            for _tool, _fields in kwargs['present_in_INFO'].items():
                colnames.append(_tool)
                tool_out = []
                for _f in _fields:
                    tool_out.append(record.INFO.get(_f))
                _single_out.append(tool_out)

        # Append fields from VEP annotation
        for _tool, idx in kwargs['vep_indexes'].items():
            colnames.append(_tool)
            # If tool has multiple fields
            # and at least 1 is within VEP annot
            # and other in the INFO annot.
            if _tool in kwargs['present_in_INFO'].keys():
     
                scores_so_far = scores[key]
                _updated = []
                for _v in scores_so_far:
                    if isinstance(_v, tuple) and _v[0] == _tool:
                        val = _v[1] + [vep_info[i] for i in idx]
                        _updated.append((_v[0], val))
                    else:
                        _updated.append(_v)

                scores[key] = _updated

            else:
     
                _single_out.append([vep_info[i] for i in idx])

        if kwargs['is_clinvar']:
            colnames.extend(['CLNREVSTAT', 'CLNSIG', 'CLNDISDB'])
            _single_out.extend([record.INFO.get("CLNREVSTAT"), 
                                record.INFO.get("CLNSIG"),
                                record.INFO.get("CLNDISDB")])
        
        df = pd.DataFrame( _single_out).T
        df.columns = colnames
        scores.append(df)
    vcf_data.close()

    if len(scores) != 0:
        return pd.concat(scores).set_index('index')
    


def _check_if_field_exists(field: list, available: list):
    """
    Checks whether a given field
    provided in the tools config (2nd col)
    exists in the VCF header

    :param list field: VCF field (s) from
    which predictions of a given tool should
    be extracted. If len(field) > 1,  multiple
    fields were provided in the tools config file

    :param list available: List of available fields
    to search for

    :return list: List tuples where 1st element
        is the VCF field name found and the second the
        index where it is found in the list of available
         fields

    :return list: List of VCF field name (s) that weren't
        found in the list of available fields
    """
    present, absent = [], []

    for _f in field:
        if _f in available:
            present.append((_f, available.index(_f)))
        else:
            absent.append(_f)

    return present, absent


def _select_consequence(vep_annotations: str, **kwargs):
    """
    Selects top VEP consequence block based on a
    choice provided by the user.

    :param str vep_annotations: Str with VEP
     annotations. If multiple blocks, "," splits
     them

    :return str: Top block
    """

    if "," in vep_annotations:
        
        if kwargs['select_conseq'] == "first":
            return vep_annotations.split(",")[0].split("|")
        
        elif kwargs['select_conseq'] in ["gene_body", "smallest_offset"]:
       
            try:
                hgvsc_index = kwargs['all_vep_annotations'].index('HGVSc')
            
            except ValueError:
                raise ValueError('HGVSc not present in VEP annotations.'
                                 'Can\'t select top consequence block based '
                                 'on the \'gene_body\' flag')
            
         
            if kwargs['select_conseq'] == "gene_body":
            
                for block in vep_annotations.split(","):
                    _block = block.split("|")
                
                    if _block[hgvsc_index]:
                        return _block
        
                return vep_annotations.split(",")[0].split("|")
            
            else:
                i = 0
                offset = 5000000
                for _i, block in enumerate(vep_annotations.split(",")):
                    _block = block.split("|")
                    
                    # if gene body
                    if _block[hgvsc_index]:
                    
                        _aux = re.findall(r'c.+[+-](\d+)', _block[hgvsc_index].split(" ")[0])
             
                        if len(_aux) > 0:
                            _offset = int(_aux[0])
                         
                        else:
                            _offset = 0
                        
                        # keep the smallest offset
                        if _offset < offset:
                            offset = _offset
                            i = _i
                            
                return vep_annotations.split(",")[i].split("|")             
            
    else:
        return vep_annotations.split("|")
